package org.processmining.processprofiler3d.compviz;

import java.util.Collection;
import java.util.HashMap;
import java.util.Vector;

import org.processmining.models.graphbased.directed.ContainableDirectedGraphElement;
import org.processmining.models.graphbased.directed.DirectedGraph;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Arc;
import org.processmining.models.graphbased.directed.petrinet.elements.ExpandableSubNet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;

public class PetriNetDataSource implements INetDataSource
{
	private PetrinetGraph snet;
	private HashMap<Transition, SubNetDataSource> subnets = new HashMap<Transition, SubNetDataSource>();
	private HashMap<String, Transition> transitionBySubnetId = new HashMap<String, Transition>();
	
	public PetriNetDataSource(PetrinetGraph pnet)
	{
		this.snet = pnet;

		for (Transition t : snet.getTransitions())
		{
			if (t.getAttributeMap().containsKey("PNMLSubnetId"))
			{
				transitionBySubnetId.put((String)t.getAttributeMap().get("PNMLSubnetId"), t);
			}
		}
		
		for (ExpandableSubNet esn : getGroups())
		{
			if (esn.getParent() == null)
			{
				Transition tx = getTransitionForSubnet(esn);
				subnets.put(tx, new SubNetDataSource(this, esn));
			}
		}
	}
	
	public String getNetStats()
	{
		int numPlaces = 0;
		int numTransitions = 0;
		int numArcs = 0;
		int numSubnets = 0;
		
		numPlaces = numPlaces + getPlaces().size();
		numTransitions = numTransitions + getTransitions().size();
		numArcs = numArcs + getEdges().size();
		
		for (SubNetDataSource snds : getAllSubnets())
		{
			numSubnets++;

			numPlaces = numPlaces + snds.getPlaces().size();
			numTransitions = numTransitions + snds.getTransitions().size();
			numArcs = numArcs + snds.getEdges().size();
		}
		
		return "Petri net statistics:\r\nNodes: " + (numPlaces + numTransitions) + "\r\nEdges: " + numArcs + "\r\nSubnets: " + numSubnets + "\r\n";
	}
	
	public DirectedGraph<?,?> getDirectedGraph()
	{
		return snet.getGraph();
	}
	
	public Collection<ExpandableSubNet> getGroups()
	{
		return snet.getGroups();
	}
	
	public Collection<SubNetDataSource> getSubnets()
	{
		return subnets.values();
	}
	
	public boolean hasSubnet(Transition t)
	{
		return subnets.containsKey(t);
	}
	
	public boolean hasParent()
	{
		return false;
	}
	
	public INetDataSource getParent()
	{
		return null;
	}
	
	public boolean hasParentTrans()
	{
		return false;
	}
	
	public Transition getParentTrans()
	{
		return null;
	}
	
	public Collection<Place> getPlaces() 
	{
		return snet.getPlaces();
	}

	public Collection<Transition> getTransitions() 
	{
		return snet.getTransitions();
	}
	
	public Collection<Transition> getAllTransitions()
	{
		Vector<Transition> lst = new Vector<Transition>();
		lst.addAll(snet.getTransitions());
		
		for (SubNetDataSource snds : getSubnets())
		{
			getAllTransitionsHelper(snds, lst);
		}
		
		return lst;
	}
	
	public void getAllTransitionsHelper(SubNetDataSource snds, Vector<Transition> lst)
	{
		lst.addAll(snds.getTransitions());
		
		for (SubNetDataSource snds2 : snds.getSubnets())
		{
			getAllTransitionsHelper(snds2, lst);
		}
	}

	public Collection<Transition> getPreceedingTransitions(Transition t) 
	{
		Vector<Transition> lst = new Vector<Transition>();
		
		for (Place p : getPlaces())
		{
			Collection<Arc> inEdges = getInEdges(p);
			Collection<Arc> outEdges = getOutEdges(p);
			
			if (inEdges.isEmpty() || outEdges.isEmpty())
				continue;
			
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> oEdge : outEdges)
			{
				Transition t1 = (Transition)oEdge.getTarget();
				
				if (t1.getLocalID() == t.getLocalID())
				{
					for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> iEdge : inEdges)
					{
						Transition t2 = (Transition)iEdge.getSource();
						lst.add(t2);
					}
				}
			}
		}
		
		return lst;
	}

	public Arc getArc(PetrinetNode source, PetrinetNode target)
	{
		return snet.getArc(source, target);
	}
	
	public Collection<Arc> getEdges() 
	{
		Vector<Arc> lst = new Vector<Arc>();
		
		for (ContainableDirectedGraphElement dge : snet.getEdges())
		{
			if (dge instanceof Arc)
			{
				lst.add((Arc)dge);
			}
		}
		
		return lst;
	}
	
	public Collection<Arc> getInEdges(PetrinetNode pn)
	{
		Vector<Arc> lst = new Vector<Arc>();
		
		for (ContainableDirectedGraphElement dge : snet.getInEdges(pn))
		{
			if (dge instanceof Arc)
			{
				lst.add((Arc)dge);
			}
		}
		
		return lst;
	}
	
	public Collection<Arc> getOutEdges(PetrinetNode pn)
	{
		Vector<Arc> lst = new Vector<Arc>();
		
		for (ContainableDirectedGraphElement dge : snet.getOutEdges(pn))
		{
			if (dge instanceof Arc)
			{
				lst.add((Arc)dge);
			}
		}
		
		return lst;
	}

	
	public Place getStartPlace()
	{
		for (Place p : getPlaces())
		{
			if (snet.getInEdges(p).isEmpty())
			{
				return p;
			}
		}
		return null;
	}
	
	public Place getEndPlace()
	{
		for (Place p : getPlaces())
		{
			if (snet.getOutEdges(p).isEmpty())
			{
				return p;
			}
		}
		return null;
	}
	
	public SubNetDataSource getSubnetForTransition(Transition t)
	{
		return subnets.get(t);
	}
	
	public Transition getTransitionForSubnet(ExpandableSubNet esn)
	{
		return getTransitionForSubnet(esn.getLabel());
	}
	
	public Transition getTransitionForSubnet(SubNetDataSource esn)
	{
		return getTransitionForSubnet(esn.getSubnetId());
	}
	
	public Transition getTransitionForSubnet(String esnId)
	{
		return transitionBySubnetId.get(esnId);
	}

	public boolean existsInNet(String tId) 
	{
		boolean existsInNet = false;
		for (Transition t : getTransitions())
		{
			if (t.getLocalID().toString().equals(tId))
			{
				existsInNet = true;
				break;
			}
		}
		for (SubNetDataSource esn : getSubnets())
		{
			if (esn.existsInNet(tId))
			{
				existsInNet = true;
				break;
			}
		}
		return existsInNet;
	}

	public Vector<SubNetDataSource> getAllSubnets()
	{
		int maxDepth = getMaxSubnetDepth();
		Vector<SubNetDataSource> subnets = new Vector<SubNetDataSource>();
		for (int i = 0; i <= maxDepth; i++)
		{
			getAllSubnetsAtLevel(i, subnets);
		}
		
		return subnets;
	}
	
	public void getAllSubnetsAtLevel(int targetLvl, Vector<SubNetDataSource> subnets)
	{
		getAllSubnetsAtLevel(0, targetLvl, subnets);
	}
	
	public void getAllSubnetsAtLevel(int lvl, int targetLvl, Vector<SubNetDataSource> subnets)
	{
		if (lvl > targetLvl)
		{
			return;
		}
		else if (lvl == targetLvl)
		{
			for (SubNetDataSource esn : getSubnets())
			{
				subnets.addElement(esn);
			}
		}
		else
		{
			for (SubNetDataSource esn : getSubnets())
			{
				esn.getAllSubnetsAtLevel(lvl + 1, targetLvl, subnets);
			}
		}
	}
	
	public int getMaxSubnetDepth()
	{
		int depth = 0;
		for (SubNetDataSource esn : getSubnets())
		{
			depth = Math.max(depth, esn.getMaxSubnetDepth() + 1);
		}
		
		return depth;
	}
}
