package org.processmining.processprofiler3d.compviz;

import java.awt.Dimension;
import java.awt.geom.Point2D;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.processmining.models.connections.GraphLayoutConnection;
import org.processmining.models.graphbased.ViewSpecificAttributeMap;
import org.processmining.models.graphbased.directed.DirectedGraph;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.jgraph.ProMGraphModel;
import org.processmining.models.jgraph.ProMJGraph;
import org.processmining.visualisation3d.maths.JoglVectord3;

import com.jgraph.layout.JGraphFacade;
import com.jgraph.layout.JGraphLayout;
import com.jgraph.layout.hierarchical.JGraphHierarchicalLayout;

//import com.jgraph.layout.JGraphLayout;
//import com.jgraph.layout.hierarchical.JGraphHierarchicalLayout;

public class PetrinetSimplifier 
{
	private INetDataSource dataSource;
	private GraphLayoutConnection layout;
	
	private Vector<Transition> transitions = new Vector<Transition>();
	private Vector<Place> places = new Vector<Place>();
	private Vector<PetrinetNode> hidden = new Vector<PetrinetNode>();
	
	private int edgeNum = 0;
	private Vector<PetrinetNode> edgeSources = new Vector<PetrinetNode>();
	private Vector<PetrinetNode> edgeTargets = new Vector<PetrinetNode>();
	private Vector<List<JoglVectord3>> edgePoints = new Vector<List<JoglVectord3>>();
	
	private int level = 0;
	
	public static int getMaxLevel()
	{
		return 4;
	}
	
	public PetrinetSimplifier(INetDataSource pnet, GraphLayoutConnection layout, int lvl)
	{	
		this.dataSource = pnet;
		this.layout = layout;
		this.level = lvl;

		if (layout == null)
		{
			System.out.println("PetrinetSimplifier: No layout connection provided!");
			//return;
			this.layout = createLayout(pnet);
		}
		//PetrinetNode pn;

		if (lvl == 0)
		{
			Extract(pnet, this.layout);
		}
		else if (lvl == 1)
		{
			ExtractAndSimplify(pnet, this.layout);
		}
		else if (lvl == 2)
		{
			ExtractAndSimplify2(pnet, this.layout);
		}
	}
	
	private GraphLayoutConnection createLayout(INetDataSource pnet)
	{
		DirectedGraph<?,?> graph = pnet.getDirectedGraph();
		GraphLayoutConnection layoutConnection = new GraphLayoutConnection(graph);
		ProMGraphModel model = new ProMGraphModel(graph);
		ProMJGraph jgraph;
		/*
		 * Make sure that only a single ProMJGraph is created at every time.
		 * The underlying JGrpah code cannot handle creating multiple creations at the same time.
		 */
		synchronized (this) 
		{
			jgraph = new ProMJGraph(model, new ViewSpecificAttributeMap(), layoutConnection);
		}
		
		JGraphLayout layout = getLayout(0);

		if (!layoutConnection.isLayedOut()) {

			JGraphFacade facade = new JGraphFacade(jgraph);

			facade.setOrdered(false);
			facade.setEdgePromotion(true);
			facade.setIgnoresCellsInGroups(false);
			facade.setIgnoresHiddenCells(false);
			facade.setIgnoresUnconnectedCells(false);
			facade.setDirected(true);
			facade.resetControlPoints();
			if (layout instanceof JGraphHierarchicalLayout) 
			{
				facade.run((JGraphHierarchicalLayout) layout, true);
			} 
			else 
			{
				facade.run(layout, true);
			}

			Map<?, ?> nested = facade.createNestedMap(true, true);

			jgraph.getGraphLayoutCache().edit(nested);
//			jgraph.repositionToOrigin();
			layoutConnection.setLayedOut(true);

		}

		jgraph.setUpdateLayout(layout);
		
		return layoutConnection;
	}
	
	protected JGraphLayout getLayout(int orientation) 
	{
		JGraphHierarchicalLayout layout = new JGraphHierarchicalLayout();
		layout.setDeterministic(true);
		layout.setCompactLayout(true);
		layout.setFineTuning(true);
		layout.setParallelEdgeSpacing(15);
		layout.setFixRoots(false);
		
	
		layout.setOrientation(orientation);

		return layout;
	}

	private void Extract(INetDataSource pnet, GraphLayoutConnection layout) 
	{
		for (Place p : pnet.getPlaces()) 
		{
			places.add(p);
		}
		
		for (Transition t : pnet.getTransitions()) 
		{
			transitions.add(t);
			
//			if (t.getAttributeMap().containsKey("PNMLSubnetId"))
//			{
//				System.out.println("Has subprocess!");
//			}
		}
		
//		for (ExpandableSubNet esn : pnet.getGroups())
//		{
//			System.out.println("Has group!");
//		}
		
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : pnet.getEdges()) 
		{
			List<Point2D> lst = layout.getEdgePoints(edge);
			Vector<JoglVectord3> lstv = new Vector<JoglVectord3>();
			for (Point2D p : lst)
			{
				lstv.add(new JoglVectord3(-p.getX(), 0, -p.getY()));
			}
			
			PetrinetNode source = edge.getSource();
			PetrinetNode target = edge.getTarget();
			
			edgeSources.add(source);
			edgeTargets.add(target);
			edgePoints.add(lstv);
			edgeNum++;			
		}
	}

	private void ExtractAndSimplify(INetDataSource pnet, GraphLayoutConnection layout) 
	{
		for (Place p : pnet.getPlaces()) 
		{
			if (!(pnet.getInEdges(p).size() == 1 && pnet.getOutEdges(p).size() == 1))
			{
				places.add(p);	
			}
			else
			{
				hidden.add(p);
				
				PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> inEdge = pnet.getInEdges(p).iterator().next();
				PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> outEdge = pnet.getOutEdges(p).iterator().next();
				
				//add composite edge
				edgeSources.add(inEdge.getSource());
				edgeTargets.add(outEdge.getTarget());
				
				Vector<Point2D> pts = new Vector<Point2D>();
				pts.addAll(layout.getEdgePoints(inEdge));
				pts.add(layout.getPosition(p));
				pts.addAll(layout.getEdgePoints(outEdge));
				
				Vector<JoglVectord3> lstv = new Vector<JoglVectord3>();
				for (Point2D pt : pts)
				{
					lstv.add(new JoglVectord3(-pt.getX(), 0, -pt.getY()));
				}
				
				edgePoints.add(lstv);
				edgeNum++;
			}
		}
		
		for (Transition t : pnet.getTransitions()) 
		{
			transitions.add(t);
		}
		
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : pnet.getEdges()) 
		{
			PetrinetNode source = edge.getSource();
			PetrinetNode target = edge.getTarget();
			
			if (!(hidden.contains(source) || hidden.contains(target)))
			{
				List<Point2D> lst = layout.getEdgePoints(edge);
				Vector<JoglVectord3> lstv = new Vector<JoglVectord3>();
				for (Point2D p : lst)
				{
					lstv.add(new JoglVectord3(-p.getX(), 0, -p.getY()));
				}
					
				edgeSources.add(source);
				edgeTargets.add(target);
				edgePoints.add(lstv);
				edgeNum++;
			}
		}
	}

	private void ExtractAndSimplify2(INetDataSource pnet, GraphLayoutConnection layout) 
	{
		for (Place p : pnet.getPlaces()) 
		{
			if (pnet.getInEdges(p).size() == 1 && pnet.getOutEdges(p).size() == 1)
			{
				hidden.add(p);
				
				PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> inEdge = pnet.getInEdges(p).iterator().next();
				PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> outEdge = pnet.getOutEdges(p).iterator().next();
				
				//add composite edge
				edgeSources.add(inEdge.getSource());
				edgeTargets.add(outEdge.getTarget());
				
				Vector<Point2D> pts = new Vector<Point2D>();
				pts.addAll(layout.getEdgePoints(inEdge));
				pts.add(layout.getPosition(p));
				pts.addAll(layout.getEdgePoints(outEdge));
				
				Vector<JoglVectord3> lstv = new Vector<JoglVectord3>();
				for (Point2D pt : pts)
				{
					lstv.add(new JoglVectord3(-pt.getX(), 0, -pt.getY()));
				}
				
				edgePoints.add(lstv);
				edgeNum++;	
			}
			else if (pnet.getInEdges(p).size() == 1 && pnet.getOutEdges(p).size() > 1)
			{
				hidden.add(p);
				
				PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> inEdge = pnet.getInEdges(p).iterator().next();
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> outEdge : pnet.getOutEdges(p))
				{
					//add composite edge
					edgeSources.add(inEdge.getSource());
					edgeTargets.add(outEdge.getTarget());
					
					Vector<Point2D> pts = new Vector<Point2D>();
					pts.addAll(layout.getEdgePoints(inEdge));
					pts.add(layout.getPosition(p));
					pts.addAll(layout.getEdgePoints(outEdge));
					
					Vector<JoglVectord3> lstv = new Vector<JoglVectord3>();
					for (Point2D pt : pts)
					{
						lstv.add(new JoglVectord3(-pt.getX(), 0, -pt.getY()));
					}
					
					edgePoints.add(lstv);
					edgeNum++;
				}
			}
			else if (pnet.getOutEdges(p).size() == 1 && pnet.getInEdges(p).size() > 1)
			{
				hidden.add(p);
				
				PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> outEdge = pnet.getOutEdges(p).iterator().next();
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> inEdge : pnet.getInEdges(p))
				{
					//add composite edge
					edgeSources.add(inEdge.getSource());
					edgeTargets.add(outEdge.getTarget());
					
					Vector<Point2D> pts = new Vector<Point2D>();
					pts.addAll(layout.getEdgePoints(inEdge));
					pts.add(layout.getPosition(p));
					pts.addAll(layout.getEdgePoints(outEdge));
					
					Vector<JoglVectord3> lstv = new Vector<JoglVectord3>();
					for (Point2D pt : pts)
					{
						lstv.add(new JoglVectord3(-pt.getX(), 0, -pt.getY()));
					}
					
					edgePoints.add(lstv);
					edgeNum++;
				}
			}
			else
			{
				places.add(p);
			}
		}
		
		for (Transition t : pnet.getTransitions()) 
		{
			transitions.add(t);
		}
		
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : pnet.getEdges()) 
		{
			PetrinetNode source = edge.getSource();
			PetrinetNode target = edge.getTarget();
			
			if (!(hidden.contains(source) || hidden.contains(target)))
			{
				List<Point2D> lst = layout.getEdgePoints(edge);
				Vector<JoglVectord3> lstv = new Vector<JoglVectord3>();
				for (Point2D pt : lst)
				{
					lstv.add(new JoglVectord3(-pt.getX(), 0, -pt.getY()));
				}
					
				edgeSources.add(source);
				edgeTargets.add(target);
				edgePoints.add(lstv);
				edgeNum++;
			}
		}
	}
	
	public Collection<Place> getPlaces()
	{
		return places;
	}
	
	public Collection<Transition> getTransitions()
	{
		return transitions;
	}
	
	public JoglVectord3 getPosition(PetrinetNode pn)
	{
		if (layout != null)
		{
			Point2D pt = layout.getPosition(pn);
			Dimension dm = layout.getSize(pn);
			if (pt != null && dm != null)
			{
				return new JoglVectord3(-pt.getX() - (dm.getWidth() * 0.5), 0 , -pt.getY() - (dm.getHeight() * 0.5));
			}
			else
			{
				System.out.println("Layout does not contain: " + pn.getLabel());
			}	
		}
		return new JoglVectord3(0,0,0);
	}
	
	public Dimension getSize(PetrinetNode pn)
	{
		if (layout != null)
		{
			Dimension dm = layout.getSize(pn);
			if (dm != null)
			{
				return dm;
			}
			else
			{
				System.out.println("Layout does not contain size for: " + pn.getLabel());
			}
		}
		return new Dimension(1,1);
	}
	
	public int getNumEdges()
	{
		return edgeNum;
	}
	
	public PetrinetNode getEdgeSource(int i)
	{
		return edgeSources.get(i);
	}
	
	public PetrinetNode getEdgeTarget(int i)
	{
		return edgeTargets.get(i);
	}
	
	public List<JoglVectord3> getEdgePoints(int i)
	{
		return edgePoints.get(i);
	}
	
	public void setEdgePoints(int i, List<JoglVectord3> lst)
	{
		edgePoints.set(i, lst);
	}
}
