package org.processmining.processprofiler3d.compviz;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Vector;
import java.util.concurrent.TimeUnit;

import org.apache.commons.math3.stat.inference.TestUtils;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.processprofiler3d.maths.stats.LeveneTest;
import org.processmining.processprofiler3d.maths.stats.ShapiroWilkTest;
import org.processmining.processprofiler3d.maths.stats.ShapiroWilkTestResult;
import org.processmining.processprofiler3d.psen.datastructure.IPerformanceStatisticsEnrichedNet;
import org.processmining.processprofiler3d.psen.replayanalysis.TimeSpan;

public class SignificanceProviderTransition 
{
	private PetriNetDataSource pnet;
	private IPerformanceStatisticsEnrichedNet psnet;
	private HashSet<String> visClassifiers;
	
	public double pLevel = 0.05;
	
	public void setVisibleClassifiers(HashSet<String> vc)
	{
		visClassifiers = vc;
	}
	
	public SignificanceProviderTransition(PetriNetDataSource pnet, IPerformanceStatisticsEnrichedNet psnet)
	{
		this.pnet = pnet;
		this.psnet = psnet;
	}
	
	public HashMap<String, Boolean> getMap(String label1)
	{
		Date startTime = new Date();
		HashMap<String, Boolean> map = new HashMap<String, Boolean>();
		
		for (String tId : psnet.getTransitions())
		{
			List<double[]> classes = new ArrayList<double[]>();
			
			for (String classifier : psnet.getClassifierValues())
			{
				if (visClassifiers.contains(classifier))
				{
					FillClassValues(label1, tId, classes, classifier);
				}
			}
			
			map.put(tId, CalcSignificance(label1, tId, classes));
		}
		
		Date endTime = new Date();
		TimeSpan ts = new TimeSpan(startTime, endTime);
		
		System.out.println("SignificanceMap generated (" + ts.ToString(TimeUnit.MILLISECONDS) + "ms).");
		
		return map;
	}

	public void FillClassValues(String label1, String tId, List<double[]> classes, String classifier) 
	{
		Vector<Double> valList = psnet.getValues(tId, label1, classifier);
		double[] valArray = new double[valList.size()];
		int i = 0;
		for (Double d : valList)
		{
			valArray[i] = d.doubleValue();
			i++;
		}
		classes.add(valArray);
	}

	public boolean CalcSignificance(String label1, String tId, List<double[]> classes) 
	{
		Vector<double[]> validClasses = new Vector<double[]>();
		String tName = getTransitionName(pnet, tId);
		
		try
		{
			//test assumptions
			boolean bAssumptionsMet = true;
			
			//still need levene's test
			LeveneTest lt = new LeveneTest();
			if (!lt.test(classes, 0.1))
			{
				bAssumptionsMet = false;
				System.out.println("Levene's test failed: " + tName + " - " + label1);
			}
			else
			{
				System.out.println("Levene's test passed: " + tName + " - " + label1);
			}
			
			//shapiro-wilk test
			int i = 1;
			for (double[] ds : classes)
			{
				boolean bAssumptionsMetClass = true;
				ShapiroWilkTestResult swres = ShapiroWilkTest.test(ds);
				if (swres.pValue > 0.1) //relaxed significance requirements because ANOVA should be robust against violation of normality assumption
				{
					//fails assumption of normality of distributions
					bAssumptionsMetClass = false;
					//bAssumptionsMet = false;
					if (swres.pValue < 1.0)
					{
						System.out.println("Assumption of normality violated: " + tName + " - " + label1 + "[" + i + "] (p=" + swres.pValue + ")");
					}
					else
					{
						String reason = "Test failed";
						if (swres.errorCode == swres.ERROR_N_LESS_THAN_3)
						{
							reason = "Test failed: n < 3";
						}
						else if (swres.errorCode == swres.ERROR_N_GREATER_THAN_5000)
						{
							reason = "Test failed: n > 5000";
						}
						else if (swres.errorCode == swres.ERROR_UNKNOWN1)
						{
							reason = "Test failed: unknown 1";
						}
						else if (swres.errorCode == swres.ERROR_UNKNOWN2)
						{
							reason = "Test failed: unknown 2";
						}
						else if (swres.errorCode == swres.ERROR_UNKNOWN3)
						{
							reason = "Test failed: unknown 3";
						}
						else if (swres.errorCode == swres.ERROR_ZERO_RANGE)
						{
							reason = "Test failed: zero range";
						}
						else if (swres.errorCode == swres.ERROR_DATA_NOT_SORTED)
						{
							reason = "Test failed: data not sorted";
						}
						System.out.println("Assumption of normality violated: " + tName + " - " + label1 + "[" + i + "] (" + reason + ")");
					}
				}
				if (bAssumptionsMetClass)
				{
					validClasses.add(ds);
				}
				
				i++;
			}
			
			if (bAssumptionsMet && validClasses.size() > 1)
			{
				boolean bSigDiff = TestUtils.oneWayAnovaTest(validClasses, pLevel); // true means reject null hypothesis
				if (bSigDiff)
				{
					System.out.println("Significant difference: " + tName + " - " + label1);
				}
				return bSigDiff;
			}
		}
		catch (Exception ex)
		{
			
		}
		
		return false;
	}
	
	private String getTransitionName(INetDataSource nds, String tId)
	{
		for (Transition t : nds.getTransitions())
		{
			if (t.getLocalID().toString().equals(tId))
			{
				return t.getLabel();
			}
			
			for (SubNetDataSource snds : nds.getSubnets())
			{
				String ts = getTransitionName(snds, tId);
				if (!ts.equals(tId))
				{
					return ts;
				}
			}
		}
		
		return tId;
	}
}
