package org.processmining.processprofiler3d.compviz;

import java.util.Collection;
import java.util.HashMap;
import java.util.Vector;

import org.processmining.models.graphbased.directed.ContainableDirectedGraphElement;
import org.processmining.models.graphbased.directed.DirectedGraph;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Arc;
import org.processmining.models.graphbased.directed.petrinet.elements.ExpandableSubNet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;

public class SubNetDataSource implements INetDataSource
{
	private INetDataSource parent = null;
	public ExpandableSubNet snet;

	private Vector<Transition> transitions = new Vector<Transition>();
	private Vector<Place> places = new Vector<Place>();
	private Vector<ExpandableSubNet> groups = new Vector<ExpandableSubNet>();
	private HashMap<Transition, SubNetDataSource> subnets = new HashMap<Transition, SubNetDataSource>();
	private HashMap<String, Transition> transitionBySubnetId = new HashMap<String, Transition>();
	private Vector<Arc> edges = new Vector<Arc>();
	
	public SubNetDataSource(INetDataSource parent, ExpandableSubNet snet)
	{
		this.snet = snet;
		this.parent = parent;
		
		for (ContainableDirectedGraphElement dge : snet.getChildren())
		{
			if (dge instanceof Transition)
			{
				Transition t = (Transition)dge;
				transitions.add(t);
				
				if (t.getAttributeMap().containsKey("PNMLSubnetId"))
				{
					transitionBySubnetId.put((String)t.getAttributeMap().get("PNMLSubnetId"), t);
				}
			}
			else if (dge instanceof Place)
			{
				places.add((Place)dge);
			}
			else if (dge instanceof ExpandableSubNet)
			{
				groups.add((ExpandableSubNet)dge);
			}
			else if (dge instanceof Arc)
			{
				edges.add((Arc)dge);
			}
		}
		
		for (ExpandableSubNet esn : getGroups())
		{
			if (esn.getParent() == snet)
			{
				Transition tx = getTransitionForSubnet(esn);
				subnets.put(tx, new SubNetDataSource(this, esn));
			}
		}
	}
	
	public boolean hasParent()
	{
		return true;
	}
	
	public INetDataSource getParent()
	{
		return parent;
	}
	
	public boolean hasParentTrans()
	{
		return true;
	}
	
	public Transition getParentTrans()
	{
		return parent.getTransitionForSubnet(this);
	}
	
	public DirectedGraph<?,?> getDirectedGraph()
	{
		return snet.getGraph();
	}
	
	public String getSubnetId()
	{
		return snet.getLabel();
	}
	
	public boolean hasSubnet(Transition t)
	{
		return subnets.containsKey(t);
	}
	
	public Collection<ExpandableSubNet> getGroups()
	{
		return groups;
	}
	
	public Collection<SubNetDataSource> getSubnets()
	{
		return subnets.values();
	}
	
	public Collection<Place> getPlaces() 
	{
		return places;
	}

	public Collection<Transition> getTransitions() 
	{
		return transitions;
	}

	public Collection<Transition> getPreceedingTransitions(Transition t) 
	{
		Vector<Transition> lst = new Vector<Transition>();
		
		for (Place p : places)
		{
			Collection<Arc> inEdges = getInEdges(p);
			Collection<Arc> outEdges = getOutEdges(p);
			
			if (inEdges.isEmpty() || outEdges.isEmpty())
				continue;
			
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> oEdge : outEdges)
			{
				Transition t1 = (Transition)oEdge.getTarget();
				
				if (t1.getLocalID() == t.getLocalID())
				{
					for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> iEdge : inEdges)
					{
						Transition t2 = (Transition)iEdge.getSource();
						lst.add(t2);
					}
				}
			}
		}
		
		return lst;
	}

	public Arc getArc(PetrinetNode source, PetrinetNode target)
	{
		for (Arc a : edges)
		{
			if (a.getSource() == source && a.getTarget() == target)
			{
				return a;
			}
		}
		return null;
	}
	
	public Collection<Arc> getEdges() 
	{
		return edges;
	}
	
	public Collection<Arc> getInEdges(PetrinetNode pn)
	{
		Vector<Arc> lst = new Vector<Arc>();
		
		for (Arc a : edges)
		{
			if (a.getTarget() == pn)
			{
				lst.add(a);
			}
		}
		
		return lst;
	}
	
	public Collection<Arc> getOutEdges(PetrinetNode pn)
	{
		Vector<Arc> lst = new Vector<Arc>();
		
		for (Arc a : edges)
		{
			if (a.getSource() == pn)
			{
				lst.add(a);
			}
		}
		
		return lst;
	}

	
	public Place getStartPlace()
	{
		for (Place p : places)
		{
			if (getInEdges(p).isEmpty())
			{
				return p;
			}
		}
		return null;
	}
	
	public Place getEndPlace()
	{
		for (Place p : places)
		{
			if (getOutEdges(p).isEmpty())
			{
				return p;
			}
		}
		return null;
	}
	
	public SubNetDataSource getSubnetForTransition(Transition t)
	{
		return subnets.get(t);
	}
	
	public Transition getTransitionForSubnet(ExpandableSubNet esn)
	{
		return getTransitionForSubnet(esn.getLabel());
	}
	
	public Transition getTransitionForSubnet(SubNetDataSource esn)
	{
		return getTransitionForSubnet(esn.getSubnetId());
	}
	
	public Transition getTransitionForSubnet(String esnId)
	{
		return transitionBySubnetId.get(esnId);
	}

	public boolean existsInNet(String tId) 
	{
		boolean existsInNet = false;
		for (Transition t : getTransitions())
		{
			if (t.getLocalID().toString().equals(tId))
			{
				existsInNet = true;
				break;
			}
		}
		for (SubNetDataSource esn : getSubnets())
		{
			if (esn.existsInNet(tId))
			{
				existsInNet = true;
				break;
			}
		}
		return existsInNet;
	}

	public void getAllSubnetsAtLevel(int targetLvl, Vector<SubNetDataSource> subnets)
	{
		getAllSubnetsAtLevel(0, targetLvl, subnets);
	}
	
	public void getAllSubnetsAtLevel(int lvl, int targetLvl, Vector<SubNetDataSource> subnets)
	{
		if (lvl > targetLvl)
		{
			return;
		}
		else if (lvl == targetLvl)
		{
			for (SubNetDataSource esn : getSubnets())
			{
				subnets.addElement(esn);
			}
		}
		else
		{
			for (SubNetDataSource esn : getSubnets())
			{
				esn.getAllSubnetsAtLevel(lvl + 1, targetLvl, subnets);
			}
		}
	}
	
	public int getMaxSubnetDepth()
	{
		int depth = 0;
		for (SubNetDataSource esn : getSubnets())
		{
			depth = Math.max(depth, esn.getMaxSubnetDepth() + 1);
		}
		
		return depth;
	}
}
