package org.processmining.processprofiler3d.placealignment;


import java.awt.Dimension;
import java.awt.event.KeyEvent;
import java.util.Vector;

import javax.swing.BoxLayout;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTable;

import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.processprofiler3d.compviz.PetriNetDataSource;
import org.processmining.processprofiler3d.psen.AlignedLog;
import org.processmining.processprofiler3d.psen.LogAnalyser;
import org.processmining.processprofiler3d.psen.datastructure.IPerformanceStatisticsEnrichedNet;
import org.processmining.processprofiler3d.psen.gui.SlickerOptionPane;
import org.processmining.processprofiler3d.psen.plugins.PP3DPluginBase;
import org.processmining.processprofiler3d.psen.preprocess.PlaceAlignmentAnalysisReport;


/*
 * This is the entry class of the comparative visualisation plug-in.
 * */
public class PlaceAlignmentAnalysis extends PP3DPluginBase
{
	@Plugin(
		name = "Process Profiler 3D: Analyse Place Alignment", 
		parameterLabels = { "Petri Net", "Event Log"}, 
		returnLabels = { "Performance Statistics Enriched Net", "ConvertedLog" }, 
		returnTypes = { IPerformanceStatisticsEnrichedNet.class, XLog.class}, 
		userAccessible = true, 
		help = "This plug-in analyses the quality of alignment of places in a process model to a process log."
	)
	@UITopiaVariant(
		affiliation = "Queensland University of Technology", 
		author = "Erik Poppe", 
		email = "erik.poppe@qut.edu.au"
	)
	
	public Object[] replayALog(final UIPluginContext context, Petrinet hierarchicalNet, XLog log) 
	{
		LogAnalyser la = new LogAnalyser(log);

		Object[] results = null;
		Vector<PlaceAlignmentInfo> paiLst = null;

		//check if trace names are unique
		if (!la.areTraceNamesUnique())
		{
			//SlickerOptionPane.showMessageDialog(new JPanel(), "Trace names in log are not unique! Cannot analyse.");
			ProMUIHelper.showErrorMessage("Trace names in log are not unique! Cannot analyse.", "Error");
			System.out.println("Trace names in log are not unique! Cannot analyse.");
		}
		else
		{
			//get initial markings
			Marking initMarking = getInitialMarking(context, hierarchicalNet);
			if (initMarking == null)
			{
				//SlickerOptionPane.showMessageDialog(new JPanel(), "Initial Marking for petrinet could not be found or created! Cannot simulate replay.");
				ProMUIHelper.showErrorMessage("Initial Marking for petrinet could not be found or created! Cannot simulate replay.", "Error");
				System.out.println("Initial Marking for petrinet could not be found or created! Cannot simulate replay.");

				context.getFutureResult(0).cancel(true);
				return null;
			}
			
			if (initMarking.isEmpty())
			{
				//SlickerOptionPane.showMessageDialog(new JPanel(), "Initial Marking for petrinet is empty! Cannot simulate replay.");
				ProMUIHelper.showErrorMessage("Initial Marking for petrinet is empty! Cannot simulate replay.", "Error");
				System.out.println("Initial Marking for petrinet is empty! Cannot simulate replay.");

				context.getFutureResult(0).cancel(true);
				return null;
			}
			
			Marking finalMarking = getFinalMarking(context, hierarchicalNet);
			
			if (finalMarking == null)
			{
				//SlickerOptionPane.showMessageDialog(new JPanel(), "Final Marking for petrinet could not be found or created! Cannot simulate replay.");
				ProMUIHelper.showErrorMessage("Final Marking for petrinet could not be found or created! Cannot simulate replay.", "Error");
				System.out.println("Final Marking for petrinet could not be found or created! Cannot simulate replay.");

				context.getFutureResult(0).cancel(true);
				return null;
			}
			
			if (finalMarking.isEmpty())
			{
				//SlickerOptionPane.showMessageDialog(new JPanel(), "Final Marking for petrinet is empty! Cannot simulate replay.");
				ProMUIHelper.showErrorMessage("Final Marking for petrinet is empty! Cannot simulate replay.", "Error");
				System.out.println("Final Marking for petrinet is empty! Cannot simulate replay.");

				context.getFutureResult(0).cancel(true);
				return null;
			}
			
			//if log is already annotated with alignment information, no need to align log again for replay
			if (AlignedLog.CheckLogIsAligned(la))
			{
				paiLst = PlaceAlignmentCalculator.Analyse(new PetriNetDataSource(hierarchicalNet), log, initMarking, finalMarking);
				
				results = new Object[2];
				results[0] = null;
				results[1] = null;
			}
			//if log is not annotated with alignment information, align log for replay first
			else
			{

				SlickerOptionPane.showMessageDialog(context.getGlobalContext().getUI(), "Log has no alignment annotation! Cannot analyse place alignment.");
				//ProMUIHelper.showErrorMessage("Log has no alignment annotation! Cannot analyse place alignment.", "Error");
				
				results = new Object[2];
				results[0] = null;
				results[1] = null;
				
			}
		}
		
		
		if (results == null)
		{
			context.getFutureResult(0).cancel(true);
		}
		else
		{

			if (paiLst != null)
			{
				JComponent listPane = CreateAnalysisSummary(paiLst);
				
				//InteractionResult ir = context.showConfiguration("Results", listPane);
				PlaceAlignmentAnalysisReport paar = new PlaceAlignmentAnalysisReport("Place Alignment", listPane);
				paar.ShowDialog(context);
			}
			
			//String orgLogName = XConceptExtension.instance().extractName(log);
			//context.getFutureResult(1).setLabel(orgLogName + " Annotated");
		}
		
		return results;
	}
	
	public JComponent CreateAnalysisSummary(Vector<PlaceAlignmentInfo> paiLst)
	{
		JTabbedPane tabbedPane = new JTabbedPane();
		//ImageIcon icon = createImageIcon("images/middle.gif");

//		JComponent panel1 = CreateScrollableTab(CreateAnalysisSummaryPanelAbs(paiLst));
//		tabbedPane.addTab("Abs", null, panel1,
//		                  "Absolute metrics");
//		tabbedPane.setMnemonicAt(0, KeyEvent.VK_1);
//
//		JComponent panel2 = CreateScrollableTab(CreateAnalysisSummaryPanelRel(paiLst));
//		tabbedPane.addTab("Rel", null, panel2,
//		                  "Relative metrics");
//		tabbedPane.setMnemonicAt(1, KeyEvent.VK_2);
//
//		JComponent panel8 = CreateScrollableTab(CreateAnalysisSummaryPanelQual(paiLst));
//		tabbedPane.addTab("Qual", null, panel8,
//		                  "Quality metrics");
//		tabbedPane.setMnemonicAt(2, KeyEvent.VK_3);
//
//		JComponent panel3 = CreateScrollableTab(CreateAnalysisSummaryPanelProb(paiLst));
//		tabbedPane.addTab("Prob", null, panel3,
//		                  "Incorrect ordering probabilities");
//		tabbedPane.setMnemonicAt(3, KeyEvent.VK_4);
//
//		JComponent panel4 = CreateScrollableTab(CreateAnalysisSummaryPanelDetails(paiLst));
//		tabbedPane.addTab("Detail", null, panel4,
//		                  "Arc absolute metrics");
//		tabbedPane.setMnemonicAt(4, KeyEvent.VK_5);
//
//		JComponent panel5 = CreateScrollableTab(CreateAnalysisSummaryPanelDetailsRel(paiLst));
//		tabbedPane.addTab("DetailRel", null, panel5,
//		                  "Arc relative metrics");
//		tabbedPane.setMnemonicAt(5, KeyEvent.VK_6);
//
//		JComponent panel6 = CreateScrollableTab(CreateAnalysisSummaryPanelDetailsQual(paiLst));
//		tabbedPane.addTab("DetailQual", null, panel6,
//		                  "Arc quality metrics");
//		tabbedPane.setMnemonicAt(6, KeyEvent.VK_7);
//
//		JComponent panel7 = CreateScrollableTab(CreateAnalysisSummaryPanelDetailsProb(paiLst));
//		tabbedPane.addTab("DetailProb", null, panel7,
//		                  "Arc incorrect ordering probabilities");
//		tabbedPane.setMnemonicAt(7, KeyEvent.VK_8);

		JComponent panel8 = CreateScrollableTab(CreateAnalysisSummaryPanelQual(paiLst));
		tabbedPane.addTab("Qual", null, panel8,
		                  "Quality metrics");
		tabbedPane.setMnemonicAt(0, KeyEvent.VK_1);

		JComponent panel3 = CreateScrollableTab(CreateAnalysisSummaryPanelProb(paiLst));
		tabbedPane.addTab("Prob", null, panel3,
		                  "Incorrect ordering probabilities");
		tabbedPane.setMnemonicAt(1, KeyEvent.VK_2);

		JComponent panel6 = CreateScrollableTab(CreateAnalysisSummaryPanelDetailsQual(paiLst));
		tabbedPane.addTab("DetailQual", null, panel6,
		                  "Arc quality metrics");
		tabbedPane.setMnemonicAt(2, KeyEvent.VK_3);

		JComponent panel7 = CreateScrollableTab(CreateAnalysisSummaryPanelDetailsProb(paiLst));
		tabbedPane.addTab("DetailProb", null, panel7,
		                  "Arc incorrect ordering probabilities");
		tabbedPane.setMnemonicAt(3, KeyEvent.VK_4);
		
		return tabbedPane;
	}
	
	private static JComponent CreateScrollableTab(JPanel listPane)
	{
		JPanel listPane2 = new JPanel();
		listPane2.setPreferredSize(new Dimension(500,400));
		listPane2.setLayout(new BoxLayout(listPane2, BoxLayout.PAGE_AXIS));

		JScrollPane jsp = new JScrollPane(listPane);
		//jsp.add(listPane);
		listPane2.add(jsp);
		
		return listPane2;
	}
	
	private static JPanel CreateAnalysisSummaryPanelAbs(Vector<PlaceAlignmentInfo> paiLst)
	{
		JPanel listPane = new JPanel();
		listPane.setLayout(new BoxLayout(listPane, BoxLayout.PAGE_AXIS));

		PlaceAlignmentTableModel patm = new PlaceAlignmentTableModel(paiLst);
		
		listPane.add(new JLabel("Total Case #: " + patm.getNumCases()));
		
		JTable table = new JTable(patm);
		
		for (int i = 0; i < patm.getColumnCount(); i++)
		{
			table.getColumnModel().getColumn(i).setCellRenderer(new PlaceAlignmentCellRenderer());
		}
		
		//table.setTableHeader(new JTableHeader());
		listPane.add(table.getTableHeader());
		listPane.add(table);
		
		return listPane;
	}
	
	private static JPanel CreateAnalysisSummaryPanelRel(Vector<PlaceAlignmentInfo> paiLst)
	{
		JPanel listPane = new JPanel();
		listPane.setLayout(new BoxLayout(listPane, BoxLayout.PAGE_AXIS));

		PlaceAlignmentTableModelRel patm = new PlaceAlignmentTableModelRel(paiLst);
		
		listPane.add(new JLabel("Total Case #: " + patm.getNumCases()));
		
		JTable table = new JTable(patm);
		
		for (int i = 0; i < patm.getColumnCount(); i++)
		{
			table.getColumnModel().getColumn(i).setCellRenderer(new PlaceAlignmentCellRenderer());
		}
		
		//table.setTableHeader(new JTableHeader());
		listPane.add(table.getTableHeader());
		listPane.add(table);
		
		return listPane;
	}
	
	private static JPanel CreateAnalysisSummaryPanelQual(Vector<PlaceAlignmentInfo> paiLst)
	{
		JPanel listPane = new JPanel();
		listPane.setLayout(new BoxLayout(listPane, BoxLayout.PAGE_AXIS));

		PlaceAlignmentTableModelQual patm = new PlaceAlignmentTableModelQual(paiLst);
		
		listPane.add(new JLabel("Total Case #: " + patm.getNumCases()));
		
		JTable table = new JTable(patm);
		
		for (int i = 0; i < patm.getColumnCount(); i++)
		{
			table.getColumnModel().getColumn(i).setCellRenderer(new PlaceAlignmentCellRenderer());
		}
		
		//table.setTableHeader(new JTableHeader());
		listPane.add(table.getTableHeader());
		listPane.add(table);
		
		return listPane;
	}
	
	private static JPanel CreateAnalysisSummaryPanelProb(Vector<PlaceAlignmentInfo> paiLst)
	{	
		JPanel listPane = new JPanel();
		listPane.setLayout(new BoxLayout(listPane, BoxLayout.PAGE_AXIS));

		PlaceAlignmentTableModelProb patm = new PlaceAlignmentTableModelProb(paiLst);
		
		listPane.add(new JLabel("Total Case #: " + patm.getNumCases()));
		
		JTable table = new JTable(patm);
		
		for (int i = 0; i < patm.getColumnCount(); i++)
		{
			table.getColumnModel().getColumn(i).setCellRenderer(new PlaceAlignmentCellRenderer());
		}
		
		//table.setTableHeader(new JTableHeader());
		listPane.add(table.getTableHeader());
		listPane.add(table);
		
		return listPane;
	}
	
	private static JPanel CreateAnalysisSummaryPanelDetails(Vector<PlaceAlignmentInfo> paiLst)
	{
		JPanel listPane = new JPanel();
		listPane.setLayout(new BoxLayout(listPane, BoxLayout.PAGE_AXIS));

		PlaceAlignmentTableModelDetail patm = new PlaceAlignmentTableModelDetail(paiLst);
		
		listPane.add(new JLabel("Total Case #: " + patm.getNumCases()));
		
		JTable table = new JTable(patm);
		
		for (int i = 0; i < patm.getColumnCount(); i++)
		{
			table.getColumnModel().getColumn(i).setCellRenderer(new PlaceAlignmentCellRenderer());
		}
		
		listPane.add(table.getTableHeader());
		listPane.add(table);
		
		return listPane;
	}
	
	private static JPanel CreateAnalysisSummaryPanelDetailsRel(Vector<PlaceAlignmentInfo> paiLst)
	{
		JPanel listPane = new JPanel();
		listPane.setLayout(new BoxLayout(listPane, BoxLayout.PAGE_AXIS));

		PlaceAlignmentTableModelDetailRel patm = new PlaceAlignmentTableModelDetailRel(paiLst);
		
		listPane.add(new JLabel("Total Case #: " + patm.getNumCases()));
		
		JTable table = new JTable(patm);
		
		for (int i = 0; i < patm.getColumnCount(); i++)
		{
			table.getColumnModel().getColumn(i).setCellRenderer(new PlaceAlignmentCellRenderer());
		}
		
		listPane.add(table.getTableHeader());
		listPane.add(table);
		
		return listPane;
	}
	
	private static JPanel CreateAnalysisSummaryPanelDetailsQual(Vector<PlaceAlignmentInfo> paiLst)
	{
		JPanel listPane = new JPanel();
		listPane.setLayout(new BoxLayout(listPane, BoxLayout.PAGE_AXIS));

		PlaceAlignmentTableModelDetailQual patm = new PlaceAlignmentTableModelDetailQual(paiLst);
		
		listPane.add(new JLabel("Total Case #: " + patm.getNumCases()));
		
		JTable table = new JTable(patm);
		
		for (int i = 0; i < patm.getColumnCount(); i++)
		{
			table.getColumnModel().getColumn(i).setCellRenderer(new PlaceAlignmentCellRenderer());
		}
		
		listPane.add(table.getTableHeader());
		listPane.add(table);
		
		return listPane;
	}
	
	private static JPanel CreateAnalysisSummaryPanelDetailsProb(Vector<PlaceAlignmentInfo> paiLst)
	{
		JPanel listPane = new JPanel();
		listPane.setLayout(new BoxLayout(listPane, BoxLayout.PAGE_AXIS));

		PlaceAlignmentTableModelDetailProb patm = new PlaceAlignmentTableModelDetailProb(paiLst);
		
		listPane.add(new JLabel("Total Case #: " + patm.getNumCases()));
		
		JTable table = new JTable(patm);
		
		for (int i = 0; i < patm.getColumnCount(); i++)
		{
			table.getColumnModel().getColumn(i).setCellRenderer(new PlaceAlignmentCellRenderer());
		}
		
		listPane.add(table.getTableHeader());
		listPane.add(table);
		
		return listPane;
	}
}
