package org.processmining.processprofiler3d.psen.fileio;

import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;

import org.processmining.contexts.uitopia.annotations.UIImportPlugin;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.abstractplugins.AbstractImportPlugin;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.processprofiler3d.psen.connections.PSENConnection;
import org.processmining.processprofiler3d.psen.datastructure.H2MemoryDatabase;
import org.processmining.processprofiler3d.psen.datastructure.IPerformanceStatisticsEnrichedNet;
import org.processmining.processprofiler3d.psen.datastructure.PSENLabel;
import org.processmining.processprofiler3d.psen.datastructure.PerformanceStatisticsDB;
import org.processmining.processprofiler3d.psen.fileio.xml.XMLTag;
import org.processmining.processprofiler3d.psen.replayanalysis.TimeSpan;


/*
 * This is the entry class of the comparative visualisation plug-in.
 * */

@Plugin(name = "Import Performance Comparison",
parameterLabels = { "File" },
returnLabels = { "PerformanceStatisticsEnrichedNet" },
returnTypes = {IPerformanceStatisticsEnrichedNet.class})
@UIImportPlugin(description = "Performance Comparison file",
        extensions = "pcf")

public class ModelComparisonImport extends AbstractImportPlugin 
{
	@UITopiaVariant(affiliation = "Queensland University of Technology",
              author = "Erik Poppe",
              email = "erik.poppe@qut.edu.au")
	@Override
	protected IPerformanceStatisticsEnrichedNet importFromStream(final PluginContext context,
	                                    final InputStream input,
	                                    final String filename,
	                                    final long fileSizeInBytes) 
	{
		
	    // Fill in object from input
		BufferedReader br = new BufferedReader(new InputStreamReader(input));
		
		IPerformanceStatisticsEnrichedNet psen = null;
		IPerformanceStatisticsEnrichedNet wten = null;

		XMLTag tag = XMLTag.ReadFrom(br);
		if (tag.getTagName().equals("PSENData"))
		{
			H2MemoryDatabase h2db = new H2MemoryDatabase();
			
			if (h2db.isDriverAvailable())
			{
				XMLTag tag2 = XMLTag.ReadFrom(br);
				String psenName = "Performance Statistics Enriched Net";
				if (tag2.hasAttribute("name"))
				{
					psenName = tag2.getAttribute("name");
				}
				if (tag2.getTagName().equals("PSEN"))
				{
					psen = ReadPSEN(h2db, br);
				}
				
				tag2 = XMLTag.ReadFrom(br);
				if (tag2.getTagName().equals("PSENWaitingTimeTree"))
				{				
					wten = ReadWaitingTimeTree(h2db, br);
				}
				
				if (psen != null && wten != null)
				{
					context.addConnection(new PSENConnection(psen, wten));
	
					psen.update();
					wten.update();
	
					System.out.println("PSEN value count: " + psen.getValueCount());
					System.out.println("WTEN value count: " + wten.getValueCount());
				}
				
				context.getFutureResult(0).setLabel(psenName);
			}
			else
			{
				context.getFutureResult(0).cancel(true);
			}
		}

		return psen;
	}
		
	private IPerformanceStatisticsEnrichedNet ReadPSEN(H2MemoryDatabase h2db, BufferedReader br)
	{
		PerformanceStatisticsDB psen = new PerformanceStatisticsDB(h2db);
		psen.InitPSEN();
		
		XMLTag tag = XMLTag.ReadFrom(br);
		while (tag != null)
		{
			if (tag.getTagName().equals("Label"))
			{
				String labelName = tag.getAttribute("name");
				String labelNodeName = tag.getAttribute("node");
				String labelAggType = tag.getAttribute("aggtype");
				String labelType = tag.getAttribute("type");
				psen.addLabel(new PSENLabel(labelName, labelNodeName, Integer.parseInt(labelAggType), Integer.parseInt(labelType) == 1));
				ReadPSENLabel(br);
			}
			if (tag.getTagName().equals("Transition"))
			{
				String transitionName = tag.getAttribute("id");
				ReadPSENTransition(br, psen, transitionName);
				
			}
			if (tag.getTagName().equals("PSEN") && tag.getIsClosingTag())
			{				
				break;
			}
			tag = XMLTag.ReadFrom(br);
		}
		return psen;
	}
	
	private void ReadPSENLabel(BufferedReader br)
	{
		XMLTag tag = XMLTag.ReadFrom(br);
		while (tag != null)
		{
			if (tag.getTagName().equals("Label") && tag.getIsClosingTag())
			{				
				break;
			}
			tag = XMLTag.ReadFrom(br);
		}
	}
	
	private void ReadPSENTransition(BufferedReader br, IPerformanceStatisticsEnrichedNet psen, String tName)
	{
		XMLTag tag = XMLTag.ReadFrom(br);
		while (tag != null)
		{
			if (tag.getTagName().equals("LabelNode"))
			{
				String labelNodeName = tag.getAttribute("name");
				String labelName = "";
				
				for (PSENLabel l : psen.getLabels())
				{
					if (l.getNodeName().equals(labelNodeName))
					{
						labelName = l.getLabelName();
					}
				}
				
//				if (!psen.hasLabel(labelName))
//				{
//					//psen.addLabel(new PSENLabel(labelName, PSENLabel.SIMPLE_LABEL));
//				}
				if (labelName != "")
				{
					ReadPSENLabelNode(br, psen, tName, labelName);
				}
				else
				{
					ReadPSENLabelNode(br, psen, tName, labelNodeName);
				}
				
			}
			if (tag.getTagName().equals("Transition") && tag.getIsClosingTag())
			{				
				break;
			}
			tag = XMLTag.ReadFrom(br);
		}
	}
	
	private void ReadPSENLabelNode(BufferedReader br, IPerformanceStatisticsEnrichedNet psen, String tName, String lName)
	{
		XMLTag tag = XMLTag.ReadFrom(br);
		while (tag != null)
		{
			if (tag.getTagName().equals("Value"))
			{
				String traceid = tag.getAttribute("trace");
				//ReadPSENSimpleClassifier(br, psen, tName, lName, cName);
				String val = XMLTag.ReadToNextTag(br);
				double d = Double.parseDouble(val.trim());

				psen.addValue(tName, lName, traceid, d);

				XMLTag tag2 = XMLTag.ReadFrom(br);
				if (!(tag2.getTagName() == "Value" && tag2.getIsClosingTag()))
				{
					//there is something wrong here!
				}
			}
			if (tag.getTagName().equals("LabelNode") && tag.getIsClosingTag())
			{	
				break;
			}
			tag = XMLTag.ReadFrom(br);
		}
	}
	
//	private void ReadPSENSimpleClassifier(BufferedReader br, IPerformanceStatisticsEnrichedNet psen, String tName, String lName, String cName)
//	{
//		XMLTag tag = XMLTag.ReadFrom(br);
//		while (tag != null)
//		{
//			if (tag.getTagName().equals("Value"))
//			{	
//				String val = XMLTag.ReadToNextTag(br);
//				double d = Double.parseDouble(val.trim());
//
//				psen.addValue(tName, lName, cName, d);
//				//System.out.println("" + tName + ":" + lName + ":" + cName + ": " + val);
//				
//				XMLTag tag2 = XMLTag.ReadFrom(br);	
//				if (!(tag2.getTagName() == "Value" && tag2.getIsClosingTag()))
//				{
//					//there is something wrong here!
//				}
//				
//			}
//			if (tag.getTagName().equals("Classifier") && tag.getIsClosingTag())
//			{				
//				break;
//			}
//			tag = XMLTag.ReadFrom(br);
//		}
//	}
	
	private IPerformanceStatisticsEnrichedNet ReadWaitingTimeTree(H2MemoryDatabase h2db, BufferedReader br)
	{
		PerformanceStatisticsDB wten = new PerformanceStatisticsDB(h2db);
		wten.InitWTEN();
		
		XMLTag tag = XMLTag.ReadFrom(br);
		while (tag != null)
		{
			if (tag.getTagName().equals("Label"))
			{
				String labelName = tag.getAttribute("name");
				String labelNodeName = tag.getAttribute("node");
				String labelAggType = tag.getAttribute("aggtype");
				String labelType = tag.getAttribute("type");
				wten.addLabel(new PSENLabel(labelName, labelNodeName, Integer.parseInt(labelAggType), Integer.parseInt(labelType) == 1));
				ReadPSENLabel(br);
			}
			if (tag.getTagName().equals("TransitionPair"))
			{
				String transitionName = tag.getAttribute("id");
				ReadTreeTransitionPair(br, wten, transitionName);
				
			}
			if (tag.getTagName().equals("PSENWaitingTimeTree") && tag.getIsClosingTag())
			{				
				break;
			}
			tag = XMLTag.ReadFrom(br);
		}
		return wten;
	}
	
	private void ReadTreeTransitionPair(BufferedReader br, IPerformanceStatisticsEnrichedNet psen, String tName)
	{
		XMLTag tag = XMLTag.ReadFrom(br);
		while (tag != null)
		{
			if (tag.getTagName().equals("LabelNode"))
			{
				String labelNodeName = tag.getAttribute("name");
				String labelName = "";
				
				for (PSENLabel l : psen.getLabels())
				{
					if (l.getNodeName().equals(labelNodeName))
					{
						labelName = l.getLabelName();
					}
				}
				if (labelName != "")
				{
					ReadTreeSimpleLabel(br, psen, tName, labelName);
				}
				else
				{
					ReadTreeSimpleLabel(br, psen, tName, labelNodeName);
				}
				
			}
			if (tag.getTagName().equals("TransitionPair") && tag.getIsClosingTag())
			{				
				break;
			}
			tag = XMLTag.ReadFrom(br);
		}
	}
	
	private void ReadTreeSimpleLabel(BufferedReader br, IPerformanceStatisticsEnrichedNet psen, String tName, String lName)
	{
		XMLTag tag = XMLTag.ReadFrom(br);
		while (tag != null)
		{
			if (tag.getTagName().equals("Value"))
			{
				String traceid = tag.getAttribute("trace");
				//ReadTreeSimpleClassifier(br, psen, tName, lName, cName);
				String val = XMLTag.ReadToNextTag(br);
				Long lval = (long)Double.parseDouble(val.trim());

				psen.addValue(tName, lName, traceid, new TimeSpan(lval).getMilliseconds());

				XMLTag tag2 = XMLTag.ReadFrom(br);	
				if (!(tag2.getTagName() == "Value" && tag2.getIsClosingTag()))
				{
					//there is something wrong here!
				}
			}
			if (tag.getTagName().equals("LabelNode") && tag.getIsClosingTag())
			{	
				break;
			}
			tag = XMLTag.ReadFrom(br);
		}
	}
}
