package org.processmining.processprofiler3d.psen.plugins;


import java.awt.Dimension;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.models.connections.petrinets.EvClassLogPetrinetConnection;
import org.processmining.models.connections.petrinets.PNRepResultAllRequiredParamConnection;
import org.processmining.models.connections.petrinets.PNRepResultConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;
import org.processmining.plugins.petrinet.replayresult.PNRepResult;
import org.processmining.processprofiler3d.compviz.PetriNetDataSource;
import org.processmining.processprofiler3d.psen.ILogReplayer;
import org.processmining.processprofiler3d.psen.LogAnalyser;
import org.processmining.processprofiler3d.psen.MyPNLogReplayer;
import org.processmining.processprofiler3d.psen.MyPOLogReplayer;
import org.processmining.processprofiler3d.psen.PSENProvider;
import org.processmining.processprofiler3d.psen.PetriNet1toManyMapping;
import org.processmining.processprofiler3d.psen.connections.PSENConnection;
import org.processmining.processprofiler3d.psen.connections.PSENLogReplayConnection;
import org.processmining.processprofiler3d.psen.datastructure.IPerformanceStatisticsEnrichedNet;
import org.processmining.processprofiler3d.psen.gui.MarkingWarningDialog;
import org.processmining.processprofiler3d.psen.gui.SlickerOptionPane;
import org.processmining.processprofiler3d.psen.replayanalysis.LogReplayAnalyser;
import org.processmining.processprofiler3d.psen.replayanalysis.LogReplayAnalysisResults;
import org.processmining.processprofiler3d.psen.replayanalysis.ReplayToLogConverter;
import org.processmining.processprofiler3d.psen.replayanalysis.TraceReplayCollection;

import nl.tue.astar.AStarException;


/*
 * This is the entry class of the comparative visualisation plug-in.
 * */
public class ModelComparison extends PP3DPluginBase
{
	@Plugin(
		name = "Process Profiler 3D: Generate Performance Profile 1", 
		parameterLabels = { "Petri Net", "Event Log"}, 
		returnLabels = { "Performance Statistics Enriched Net", "Replay Results", "ConvertedLog" }, 
		returnTypes = { IPerformanceStatisticsEnrichedNet.class, PNRepResult.class, XLog.class}, 
		userAccessible = false, 
		help = "This plug-in compares the differences between cohorts in a process log."
	)
	@UITopiaVariant(
		affiliation = "Queensland University of Technology", 
		author = "Jingxin Xu, Erik Poppe", 
		email = "j15.xu@qut.edu.au, erik.poppe@qut.edu.au"
	)

	public Object[] replayALog(final UIPluginContext context, Petrinet hierarchicalNet, XLog log) 
	{
		LogAnalyser la = new LogAnalyser(log);
		
		//get initial markings
		Marking initMarking = getInitialMarking(context, hierarchicalNet);
		Marking finalMarking = getFinalMarking(context, hierarchicalNet);
		
		if (initMarking == null)
		{
			//SlickerOptionPane.showMessageDialog(new JPanel(), "Initial Marking for petrinet could not be found or created! Cannot simulate replay.");
			SlickerOptionPane.showErrorMessage(context, "Initial Marking for petrinet could not be found or created! Cannot simulate replay.", "Error");
			System.out.println("Initial Marking for petrinet could not be found or created! Cannot simulate replay.");

			context.getFutureResult(0).cancel(true);
			return null;
		}
		
		if (finalMarking == null)
		{
			//SlickerOptionPane.showMessageDialog(new JPanel(), "Final Marking for petrinet could not be found or created! Cannot simulate replay.");
			SlickerOptionPane.showErrorMessage(context, "Final Marking for petrinet could not be found or created! Cannot simulate replay.", "Error");
			System.out.println("Final Marking for petrinet could not be found or created! Cannot simulate replay.");

			context.getFutureResult(0).cancel(true);
			return null;
		}
		
		if (initMarking.isEmpty())
		{
			//SlickerOptionPane.showMessageDialog(new JPanel(), "Initial Marking for petrinet is empty! Cannot simulate replay.");
			SlickerOptionPane.showErrorMessage(context, "Initial Marking for petrinet is empty! Cannot simulate replay.", "Error");
			System.out.println("Initial Marking for petrinet is empty! Cannot simulate replay.");

			context.getFutureResult(0).cancel(true);
			return null;
		}
		
		if (finalMarking.isEmpty())
		{
			//SlickerOptionPane.showMessageDialog(new JPanel(), "Final Marking for petrinet is empty! Cannot simulate replay.");
			SlickerOptionPane.showErrorMessage(context, "Final Marking for petrinet is empty! Cannot simulate replay.", "Error");
			System.out.println("Final Marking for petrinet is empty! Cannot simulate replay.");

			context.getFutureResult(0).cancel(true);
			return null;
		}
		
		//check net structure
		PetriNetDataSource pnds = new PetriNetDataSource(hierarchicalNet);
		ModelStructureChecker msc = new ModelStructureChecker(pnds);
		if (!msc.isStructureCompliant())
		{
			JPanel listPane = GetErrorPanel(msc);
			InteractionResult ir = context.showConfiguration("Error", listPane);
			if (ir == InteractionResult.CANCEL)
			{
				context.getFutureResult(0).cancel(true);
				return null;
			}
		}

		boolean bLogHasLifecycleData = la.HasLifecycleData();
		boolean bNetHasHierarchy = !hierarchicalNet.getGroups().isEmpty();
		
		Object[] results = null;
		if (bNetHasHierarchy)
		{
			//get a flattened petri net from the hierarchical petri net
			ModelFlattenLayer mfl = new ModelFlattenLayer(context, hierarchicalNet, initMarking, finalMarking);
			Petrinet pgflat = mfl.getNet();
			//PetriNet1to1Mapping map = mfl.getMapping();
			initMarking = mfl.getInitMarking();
			finalMarking = mfl.getFinalMarking();
			
			if (bLogHasLifecycleData)
			{
				//get a lifecycle expanded petri net from the flattened petri net
				ModelLifecycleExpansionLayer mlel = new ModelLifecycleExpansionLayer(context, pgflat, initMarking, finalMarking, log);
				Petrinet pgexp = mlel.getNet();
				PetriNet1toManyMapping map2 = mlel.getMapping();
				initMarking = mlel.getInitMarking();
				finalMarking = mlel.getFinalMarking();
				
				results = ReplayOnHierarchicalExpandedNet(context, hierarchicalNet, pgflat, pgexp, map2, initMarking, log);
			}
			else
			{
				results = ReplayOnHierarchicalNet(context, hierarchicalNet, pgflat, initMarking, log);
			}
		}
		else
		{
			if (bLogHasLifecycleData)
			{
				//get a lifecycle expanded petri net from the flattened petri net
				ModelLifecycleExpansionLayer mlel = new ModelLifecycleExpansionLayer(context, hierarchicalNet, initMarking, finalMarking, log);
				Petrinet pgexp = mlel.getNet();
				PetriNet1toManyMapping map2 = mlel.getMapping();
				initMarking = mlel.getInitMarking();
				finalMarking = mlel.getFinalMarking();
				
				results = ReplayOnExpandedNet(context, hierarchicalNet, pgexp, map2, initMarking, log);
			}
			else
			{
				results = ReplayOnNet(context, hierarchicalNet, initMarking, log);
			}
		}
		
		if (results == null)
		{
			SlickerOptionPane.showErrorMessage(context, "Alignment of log and petrinet failed!", "Error");
			context.getFutureResult(0).cancel(true);
		}
		return results;
	}

	private JPanel GetErrorPanel(ModelStructureChecker msc)
	{
		JPanel listPane = new JPanel();
		listPane.setLayout(new BoxLayout(listPane, BoxLayout.PAGE_AXIS));
		
		String errorMsg = "Petri net does not comply to structural requirements:";

		JLabel lbl = new JLabel(errorMsg);
		lbl.setSize(100, 500);
		
		listPane.add(lbl);
		
		for (String sx : msc.problemList)
		{
			errorMsg += " \r\n" + sx;
			JLabel lblC = new JLabel("- " + sx);
			listPane.add(lblC);
		}

		listPane.add(Box.createRigidArea(new Dimension(5,30)));

		JLabel lblY = new JLabel("Analysis might be flawed.");
		listPane.add(lblY);
		
		JLabel lblX = new JLabel("Do you want to continue anyway?");
		listPane.add(lblX);
		
		System.out.println(errorMsg);
		return listPane;
	}
	
	private Object[] ReplayOnNet(final UIPluginContext context, Petrinet hierarchicalNet, Marking initMarking, XLog log)
	{
		//get a replay result (either existing from previous replay or by replaying log on model)
		ReplayResultPackage rrp = getReplayResult(context, hierarchicalNet, log);
		if (rrp != null)
		{
			PetriNetDataSource pnds = new PetriNetDataSource(hierarchicalNet);
			ReplayToLogConverter rtlc = new ReplayToLogConverter(hierarchicalNet, initMarking, log, rrp.logReplayResult);
			XLog rlog = rtlc.ConvertLogReplay();
			ReplayToLogConverter.fixupLog(rlog, rrp.trans2evMap, pnds);
			
			LogReplayAnalyser replayAnalyser = new LogReplayAnalyser(pnds, false, initMarking, rlog);
			LogReplayAnalysisResults lrar = replayAnalyser.AnalyseLogReplay(context); //analyse all traces available
			TraceReplayCollection provider2 = lrar.trc;
			
			PSENLogReplayConnection lrc = new PSENLogReplayConnection(hierarchicalNet, log, provider2);
			context.addConnection(lrc);
			
			//create and fill PSEN data structure
			PSENProvider provider = new PSENProvider(context, hierarchicalNet, rlog, provider2);

			IPerformanceStatisticsEnrichedNet psen = provider.getPSEnet();
			IPerformanceStatisticsEnrichedNet wten = provider.getWaitingTimeTree();
			
			if (psen != null && wten != null)
			{
				//add waitingtimeinfoconnection to context
				context.addConnection(new PSENConnection(psen, wten));
	
				System.out.println("PSEN value count: " + psen.getValueCount());
				System.out.println("WTEN value count: " + wten.getValueCount());
				
				Object[] objs = new Object[3];
				objs[0] = provider.getPSEnet();
				objs[1] = rrp.logReplayResult;
				objs[2] = rlog;
				
				return objs;
			}
		} 	
		return null;
	}
	
	private Object[] ReplayOnExpandedNet(final UIPluginContext context, Petrinet hierarchicalNet, Petrinet pgexp, PetriNet1toManyMapping map2, Marking initMarking, XLog log)
	{
		//get a replay result (either existing from previous replay or by replaying log on model)
		ReplayResultPackage rrp = getReplayResult(context, pgexp, log);
		if (rrp != null)
		{
			PetriNetDataSource pnds = new PetriNetDataSource(pgexp);
			ReplayToLogConverter rtlc = new ReplayToLogConverter(pgexp, initMarking, log, rrp.logReplayResult);
			XLog rlog = rtlc.ConvertLogReplay();
			
			LogReplayAnalyser replayAnalyser = new LogReplayAnalyser(pnds, true, initMarking, rlog);
			LogReplayAnalysisResults lrar = replayAnalyser.AnalyseLogReplay(context); //analyse all traces available
			TraceReplayCollection provider2 = lrar.trc;
			
			PSENLogReplayConnection lrc = new PSENLogReplayConnection(pgexp, log, provider2);
			context.addConnection(lrc);
			
			//TraceReplayCollection trc = provider2.ProjectBack(hierarchicalNet, map2);
			XLog prlog = ReplayToLogConverter.ProjectBack(rlog, map2);
			ReplayToLogConverter.fixupLog(prlog, rrp.trans2evMap, map2, pnds);
			TraceReplayCollection trc = provider2.ProjectBack(hierarchicalNet, map2, prlog);
			
			PSENLogReplayConnection lrc3 = new PSENLogReplayConnection(hierarchicalNet, log, trc);
			context.addConnection(lrc3);
			
			//create and fill PSEN data structure
			PSENProvider provider = new PSENProvider(context, hierarchicalNet, prlog, trc);
			
			IPerformanceStatisticsEnrichedNet psen = provider.getPSEnet();
			IPerformanceStatisticsEnrichedNet wten = provider.getWaitingTimeTree();
			
			if (psen != null && wten != null)
			{
				//add waitingtimeinfoconnection to context
				context.addConnection(new PSENConnection(psen, wten));
	
				System.out.println("PSEN value count: " + psen.getValueCount());
				System.out.println("WTEN value count: " + wten.getValueCount());
				
				Object[] objs = new Object[3];
				objs[0] = provider.getPSEnet();
				objs[1] = rrp.logReplayResult;
				objs[2] = prlog;
				
				return objs;
			}
		} 	
		return null;
	}
	
	private Object[] ReplayOnHierarchicalNet(final UIPluginContext context, Petrinet hierarchicalNet, Petrinet pgflat, Marking initMarking, XLog log)
	{
		//get a replay result (either existing from previous replay or by replaying log on model)
		ReplayResultPackage rrp = getReplayResult(context, pgflat, log);
		if (rrp != null)
		{
			PetriNetDataSource pnds = new PetriNetDataSource(hierarchicalNet);
			PetriNetDataSource pndsFlat = new PetriNetDataSource(pgflat);
			ReplayToLogConverter rtlc = new ReplayToLogConverter(pgflat, initMarking, log, rrp.logReplayResult);
			XLog rlog = rtlc.ConvertLogReplay();
			ReplayToLogConverter.fixupLog(rlog, rrp.trans2evMap, pnds);
			
			LogReplayAnalyser replayAnalyser = new LogReplayAnalyser(pndsFlat, false, initMarking, rlog);
			LogReplayAnalysisResults lrar = replayAnalyser.AnalyseLogReplay(context); //analyse all traces available
			TraceReplayCollection provider2 = lrar.trc;
			
			PSENLogReplayConnection lrc = new PSENLogReplayConnection(pgflat, log, provider2);
			context.addConnection(lrc);

			PSENLogReplayConnection lrc3 = new PSENLogReplayConnection(hierarchicalNet, log, provider2);
			context.addConnection(lrc3);
			
			provider2.addSubNetEvents(pnds);
			
			//create and fill PSEN data structure
			PSENProvider provider = new PSENProvider(context, hierarchicalNet, rlog, provider2);

			IPerformanceStatisticsEnrichedNet psen = provider.getPSEnet();
			IPerformanceStatisticsEnrichedNet wten = provider.getWaitingTimeTree();
			
			if (psen != null && wten != null)
			{
				//add waitingtimeinfoconnection to context
				context.addConnection(new PSENConnection(psen, wten));
	
				System.out.println("PSEN value count: " + psen.getValueCount());
				System.out.println("WTEN value count: " + wten.getValueCount());
				
				Object[] objs = new Object[3];
				objs[0] = provider.getPSEnet();
				objs[1] = rrp.logReplayResult;
				objs[2] = rlog;
				
				return objs;
			}
		} 	
		return null;
	}
	
	private Object[] ReplayOnHierarchicalExpandedNet(final UIPluginContext context, Petrinet hierarchicalNet, Petrinet pgflat, Petrinet pgexp, PetriNet1toManyMapping map2, Marking initMarking, XLog log)
	{
		//get a replay result (either existing from previous replay or by replaying log on model)
		ReplayResultPackage rrp = getReplayResult(context, pgexp, log);
		if (rrp != null)
		{
			PetriNetDataSource pnds = new PetriNetDataSource(pgexp);
			ReplayToLogConverter rtlc = new ReplayToLogConverter(pgexp, initMarking, log, rrp.logReplayResult);
			XLog rlog = rtlc.ConvertLogReplay();
			
			LogReplayAnalyser replayAnalyser = new LogReplayAnalyser(pnds, true, initMarking, rlog);
			LogReplayAnalysisResults lrar = replayAnalyser.AnalyseLogReplay(context); //analyse all traces available
			TraceReplayCollection provider2 = lrar.trc;
			
			PSENLogReplayConnection lrc = new PSENLogReplayConnection(pgexp, log, provider2);
			context.addConnection(lrc);
			
			//TraceReplayCollection trc = provider2.ProjectBack(pgflat, map2);
			XLog prlog = ReplayToLogConverter.ProjectBack(rlog, map2);
			ReplayToLogConverter.fixupLog(prlog, rrp.trans2evMap, map2, pnds);
			TraceReplayCollection trc = provider2.ProjectBack(pgflat, map2, prlog);

			PSENLogReplayConnection lrc2 = new PSENLogReplayConnection(pgflat, log, trc);
			context.addConnection(lrc2);
			PSENLogReplayConnection lrc3 = new PSENLogReplayConnection(hierarchicalNet, log, trc);
			context.addConnection(lrc3);
			
			trc.addSubNetEvents(new PetriNetDataSource(hierarchicalNet));
			
			//create and fill PSEN data structure		
			PSENProvider provider = new PSENProvider(context, hierarchicalNet, prlog, trc);

			IPerformanceStatisticsEnrichedNet psen = provider.getPSEnet();
			IPerformanceStatisticsEnrichedNet wten = provider.getWaitingTimeTree();
			
			if (psen != null && wten != null)
			{
				//add waitingtimeinfoconnection to context
				context.addConnection(new PSENConnection(psen, wten));
	
				System.out.println("PSEN value count: " + psen.getValueCount());
				System.out.println("WTEN value count: " + wten.getValueCount());
				
				Object[] objs = new Object[3];
				objs[0] = provider.getPSEnet();
				objs[1] = rrp.logReplayResult;
				objs[2] = prlog;
				
				return objs;
			}
		} 	
		return null;
	}
	
	private ReplayResultPackage getReplayResult(final UIPluginContext context, Petrinet pgexp, XLog log)
	{
		ReplayResultPackage rrp = new ReplayResultPackage();
		
		try
		{
			PNRepResultAllRequiredParamConnection conn = context.getConnectionManager().getFirstConnection(PNRepResultAllRequiredParamConnection.class, context, pgexp, log);
			rrp.trans2evMap = conn.getObjectWithRole(PNRepResultAllRequiredParamConnection.TRANS2EVCLASSMAPPING);
			rrp.logReplayResult = conn.getObjectWithRole(PNRepResultAllRequiredParamConnection.PNREPRESULT);
			
			System.out.println("Found replay connection.");
			context.log("Found replay connectiond.");
		}
		catch (Exception exc)
		{	
			System.out.println("Replay connection not found!");
		}
		
		//try to get replay result from previous replay using connection
		if (rrp.logReplayResult == null)
		{
			rrp.logReplayResult = getReplayResultFromConnection(context, log, pgexp, rrp.logReplayResult);
		}
		if (rrp.logReplayResult == null)
		{
			boolean bUsePOReplay = false;
			if (MyPOLogReplayer.IsPartialOrderAvailable(context))
			{

				MarkingWarningDialog mwd = new MarkingWarningDialog("Would you like to use partial order alignment?", "Use Partial Order Alignment?");
				int n = mwd.ShowDialog(context);
				
				if (n == JOptionPane.YES_OPTION)
				{
					bUsePOReplay = true;
				}
			}
			
			if (bUsePOReplay)
			{
				MyPOLogReplayer replayer2 = new MyPOLogReplayer();
				try
				{
					rrp.logReplayResult = ReplayLog(context, log, pgexp, replayer2);			
					rrp.trans2evMap = replayer2.getMapping();
				}
				catch (Exception ex)
				{
					System.out.println("Could not finish replay!");
				}
			}
			else
			{
				//Replay the event log on flattened, expanded petri net
				MyPNLogReplayer replayer = new MyPNLogReplayer();
				try
				{
					rrp.logReplayResult = ReplayLog(context, log, pgexp, replayer);			
					rrp.trans2evMap = replayer.getMapping();
				}
				catch (Exception ex)
				{
					System.out.println("Could not finish replay!");
				}
			}
		}

		if (rrp.trans2evMap == null)
		{
			// check connection in order to determine whether mapping step is needed or not
			try 
			{
				// connection is found, no need for mapping step
				// connection is not found, another plugin to create such connection
				// is automatically
				// executed
				EvClassLogPetrinetConnection evmapconn = context.getConnectionManager().getFirstConnection(EvClassLogPetrinetConnection.class, context, pgexp, log);
				rrp.trans2evMap = (TransEvClassMapping) evmapconn.getObjectWithRole(EvClassLogPetrinetConnection.TRANS2EVCLASSMAPPING);
			} 
			catch (Exception e) 
			{
				//SlickerOptionPane.showMessageDialog(new JPanel(), "No mapping can be constructed between the net and the log.");
				SlickerOptionPane.showErrorMessage(context, "No mapping can be constructed between the net and the log.", "Error");
			}
		}
		
		if (rrp.logReplayResult != null && rrp.trans2evMap != null)
		{
			return rrp;
		}
		return null;
	}

	
	private PNRepResult getReplayResultFromConnection(final UIPluginContext context, XLog log, Petrinet pgexp,
			PNRepResult logReplayResult) 
	{
		try
		{
			PNRepResultConnection conn = context.getConnectionManager().getFirstConnection(PNRepResultConnection.class, context, pgexp, log);
			logReplayResult = conn.getObjectWithRole(PNRepResultConnection.PNREPRESULT);
			
			System.out.println("Found replay connection.");
			context.log("Found replay connectiond.");
		}
		catch (Exception exc)
		{	
			System.out.println("Replay connection not found!");
		}
		return logReplayResult;
	}

	private PNRepResult ReplayLog(final UIPluginContext context, XLog log, Petrinet pgexp, ILogReplayer replayer)
			throws ConnectionCannotBeObtained, AStarException 
	{
		PNRepResult logReplayResult;
		System.out.println("Starting replay");
		context.log("Starting replay...");
		
		logReplayResult = replayer.replayLog(context, pgexp, log);
		
		System.out.println("Replay completed.");
		context.log("Replay completed.");
		
		return logReplayResult;
	}	
}
