package org.processmining.processprofiler3d.psen.plugins;


import java.util.HashSet;

import javax.swing.JLabel;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.processprofiler3d.compviz.INetDataSource;
import org.processmining.processprofiler3d.compviz.PetriNetDataSource;
import org.processmining.processprofiler3d.psen.AlignedLog;
import org.processmining.processprofiler3d.psen.LogAnalyser;
import org.processmining.processprofiler3d.psen.PSENProvider;
import org.processmining.processprofiler3d.psen.PetriNet1to1Mapping;
import org.processmining.processprofiler3d.psen.PetriNet1toManyMapping;
import org.processmining.processprofiler3d.psen.connections.PSENConnection;
import org.processmining.processprofiler3d.psen.connections.PSENLogReplayConnection;
import org.processmining.processprofiler3d.psen.datastructure.IPerformanceStatisticsEnrichedNet;
import org.processmining.processprofiler3d.psen.gui.SlickerOptionPane;
import org.processmining.processprofiler3d.psen.replayanalysis.LogReplayAnalyser;
import org.processmining.processprofiler3d.psen.replayanalysis.LogReplayAnalysisResults;
import org.processmining.processprofiler3d.psen.replayanalysis.ReplayToLogConverter;
import org.processmining.processprofiler3d.psen.replayanalysis.TraceReplayCollection;


/*
 * This is the entry class of the comparative visualisation plug-in.
 * */
public class ModelComparison2 extends PP3DPluginBase
{
	@Plugin(
		name = "Process Profiler 3D: Generate Performance Profile 2", 
		parameterLabels = { "Petri Net", "Performance Aligned Event Log"}, 
		returnLabels = { "Performance Statistics Enriched Net" }, 
		returnTypes = { IPerformanceStatisticsEnrichedNet.class}, 
		userAccessible = false, 
		help = "This plug-in compares the differences between cohorts in a aligned process log."
	)
	@UITopiaVariant(
		affiliation = "Queensland University of Technology", 
		author = "Erik Poppe", 
		email = "erik.poppe@qut.edu.au"
	)

	public IPerformanceStatisticsEnrichedNet replayALog(final UIPluginContext context, Petrinet hierarchicalNet, XLog log) 
	{
		LogAnalyser la = new LogAnalyser(log);
		
		//check if log is aligned log
		if (CheckLogIsAligned(la))
		{
			//SlickerOptionPane.showMessageDialog(new JPanel(), "Log is not aligned! Cannot analyse.");
			SlickerOptionPane.showErrorMessage(context, "Log is not aligned! Cannot analyse.", "Error");
			System.out.println("Log is not aligned! Cannot analyse.");
			context.getFutureResult(0).cancel(true);
			return null;
		}
		
		//get initial markings
		Marking initMarking = getInitialMarking(context, hierarchicalNet);
		Marking finalMarking = getFinalMarking(context, hierarchicalNet);

		if (initMarking == null || finalMarking == null)
		{
			//SlickerOptionPane.showMessageDialog(new JPanel(), "A Marking for petrinet could not be found or created! Cannot simulate replay.");
			SlickerOptionPane.showErrorMessage(context, "A Marking for petrinet could not be found or created! Cannot simulate replay.", "Error");
			System.out.println("A Marking for petrinet could not be found or created! Cannot simulate replay.");

			context.getFutureResult(0).cancel(true);
			return null;
		}
		
		if (initMarking.isEmpty() || finalMarking.isEmpty())
		{
			//SlickerOptionPane.showMessageDialog(new JPanel(), "A Marking for petrinet is empty! Cannot simulate replay.");
			SlickerOptionPane.showErrorMessage(context, "A Marking for petrinet is empty! Cannot simulate replay.", "Error");
			System.out.println("A Marking for petrinet is empty! Cannot simulate replay.");

			context.getFutureResult(0).cancel(true);
			return null;
		}
		
		//check net structure
		PetriNetDataSource pnds = new PetriNetDataSource(hierarchicalNet);
		if (!CheckSubnetStructure(pnds))
		{
			System.out.println("Petri net does not comply to structural requirements! Analysis might be flawed.");
			InteractionResult ir = context.showConfiguration("Error", new JLabel("Petri net does not comply to structural requirements. Analysis results may be flawed. Do you want to continue?"));
			if (ir == InteractionResult.CANCEL)
			{
				context.getFutureResult(0).cancel(true);
				return null;
			}
		}
		
		boolean bLogHasLifecycleData = la.HasLifecycleData();
		boolean bNetHasHierarchy = !hierarchicalNet.getGroups().isEmpty();
		
		Object[] results = null;
		PSENProvider provider = null;
		if (bNetHasHierarchy)
		{
			//get a flattened petri net from the hierarchical petri net
			ModelFlattenLayer mfl = new ModelFlattenLayer(context, hierarchicalNet, initMarking, finalMarking);
			Petrinet pgflat = mfl.getNet();
			PetriNet1to1Mapping map = mfl.getMapping();
			initMarking = mfl.getInitMarking();
			finalMarking = mfl.getFinalMarking();
			
			if (bLogHasLifecycleData)
			{
				//get a lifecycle expanded petri net from the flattened petri net
				ModelLifecycleExpansionLayer mlel = new ModelLifecycleExpansionLayer(context, pgflat, initMarking, finalMarking, log, true);
				Petrinet pgexp = mlel.getNet();
				PetriNet1toManyMapping map2 = mlel.getMapping();
				initMarking = mlel.getInitMarking();
				finalMarking = mlel.getFinalMarking();
				
				provider = ReplayOnHierarchicalExpandedNet(context, hierarchicalNet, pgflat, pgexp, map2, initMarking, log);
				//provider = ReplayOnHierarchicalExpandedNet(context, hierarchicalNet, initMarking, log);
			}
			else
			{
				provider = ReplayOnHierarchicalNet(context, hierarchicalNet, pgflat, initMarking, log);
			}
		}
		else
		{
			if (bLogHasLifecycleData)
			{
				//get a lifecycle expanded petri net from the flattened petri net
				ModelLifecycleExpansionLayer mlel = new ModelLifecycleExpansionLayer(context, hierarchicalNet, initMarking, finalMarking, log, true);
				Petrinet pgexp = mlel.getNet();
				PetriNet1toManyMapping map2 = mlel.getMapping();
				initMarking = mlel.getInitMarking();
				finalMarking = mlel.getFinalMarking();
				
				provider = ReplayOnExpandedNet(context, hierarchicalNet, pgexp, map2, initMarking, log);
				//provider = ReplayOnExpandedNet(context, hierarchicalNet, initMarking, log);
			}
			else
			{
				provider = ReplayOnNet(context, hierarchicalNet, initMarking, log);
			}
		}
		
		if (provider != null)
		{
			IPerformanceStatisticsEnrichedNet psen = provider.getPSEnet();
			IPerformanceStatisticsEnrichedNet wten = provider.getWaitingTimeTree();
			
			if (psen != null && wten != null)
			{
				//add waitingtimeinfoconnection to context
				context.addConnection(new PSENConnection(psen, wten));
	
				System.out.println("PSEN value count: " + psen.getValueCount());
				System.out.println("WTEN value count: " + wten.getValueCount());
				
				results = new Object[1];
				results[0] = provider.getPSEnet();	
			}
			else
			{
				context.getFutureResult(0).cancel(true);
			}
		}	
		else
		{
			context.getFutureResult(0).cancel(true);
		}
		return (IPerformanceStatisticsEnrichedNet)results[0];
	}
	
	private boolean CheckLogIsAligned(LogAnalyser la)
	{
		HashSet<String> lst = la.calcAvailableTraceAttrs();
		return (!lst.contains(AlignedLog.ATT_NAME_IS_RELIABLE));
	}
	
	private boolean CheckSubnetStructure(INetDataSource esn)
	{
		if (CheckSubnetStructureHelper(esn))
		{
			for (INetDataSource inds : esn.getSubnets())
			{
				if (!CheckSubnetStructure(inds))
				{
					return false;
				}
			}
			return true;
		}
		return false;
	}
	
	private boolean CheckSubnetStructureHelper(INetDataSource esn)
	{
		//check if subnet has single entry and single exit (transitions)!
		Place ps = esn.getStartPlace();
		Place pe = esn.getEndPlace();
		
		int startTransitionsNum = esn.getOutEdges(ps).size();
		int endTransitionsNum = esn.getInEdges(pe).size();
		
		if (startTransitionsNum > 1 || endTransitionsNum > 1)
		{	
			return false;
		}	
		return true;
	}
	
	private PSENProvider ReplayOnNet(final UIPluginContext context, Petrinet hierarchicalNet, Marking initMarking, XLog rlog)
	{
		PetriNetDataSource pnds = new PetriNetDataSource(hierarchicalNet);
		LogReplayAnalyser replayAnalyser = new LogReplayAnalyser(pnds, false, initMarking, rlog);
		LogReplayAnalysisResults lrar = replayAnalyser.AnalyseLogReplay(context); //analyse all traces available
		TraceReplayCollection provider2 = lrar.trc;
		
		PSENLogReplayConnection lrc = new PSENLogReplayConnection(hierarchicalNet, rlog, provider2);
		context.addConnection(lrc);
		
		//create and fill PSEN data structure
		PSENProvider provider = new PSENProvider(context, hierarchicalNet, rlog, provider2);

		return provider;
	}
	
	private PSENProvider ReplayOnExpandedNet(final UIPluginContext context, Petrinet hierarchicalNet, Petrinet pgexp, PetriNet1toManyMapping map2, Marking initMarking, XLog prlog)
	{
		XLog rlog = ReplayToLogConverter.Project(prlog, map2);

		PetriNetDataSource pnds = new PetriNetDataSource(pgexp);
		LogReplayAnalyser replayAnalyser = new LogReplayAnalyser(pnds, true, initMarking, rlog);
		LogReplayAnalysisResults lrar = replayAnalyser.AnalyseLogReplay(context); //analyse all traces available
		TraceReplayCollection provider2 = lrar.trc;
		
		PSENLogReplayConnection lrc = new PSENLogReplayConnection(pgexp, prlog, provider2);
		context.addConnection(lrc);
		
		TraceReplayCollection trc = provider2.ProjectBack(hierarchicalNet, map2, prlog);
		
		PSENLogReplayConnection lrc3 = new PSENLogReplayConnection(hierarchicalNet, prlog, trc);
		context.addConnection(lrc3);
		
		//create and fill PSEN data structure
		PSENProvider provider = new PSENProvider(context, hierarchicalNet, prlog, trc);
		
		return provider;
	}
	
	private PSENProvider ReplayOnHierarchicalNet(final UIPluginContext context, Petrinet hierarchicalNet, Petrinet pgflat, Marking initMarking, XLog rlog)
	{
		PetriNetDataSource pnds = new PetriNetDataSource(pgflat);
		LogReplayAnalyser replayAnalyser = new LogReplayAnalyser(pnds, false, initMarking, rlog);
		LogReplayAnalysisResults lrar = replayAnalyser.AnalyseLogReplay(context); //analyse all traces available
		TraceReplayCollection provider2 = lrar.trc;
		
		PSENLogReplayConnection lrc = new PSENLogReplayConnection(pgflat, rlog, provider2);
		context.addConnection(lrc);

		PSENLogReplayConnection lrc3 = new PSENLogReplayConnection(hierarchicalNet, rlog, provider2);
		context.addConnection(lrc3);
		
		provider2.addSubNetEvents(new PetriNetDataSource(hierarchicalNet));
		
		//create and fill PSEN data structure
		PSENProvider provider = new PSENProvider(context, hierarchicalNet, rlog, provider2);
		
		return provider;
	}
	
	private Transition getByLabel(Petrinet pn, String l)
	{
		for (Transition t : pn.getTransitions())
		{
			if (t.getLabel().equals(l))
			{
				return t;
			}
		}
		return null;
	}
	
//	private PSENProvider ReplayOnHierarchicalExpandedNet(final UIPluginContext context, Petrinet hierarchicalNet, Marking initMarking, XLog prlog)
//	{
//		//XLog rlog = ReplayToLogConverter.Project(prlog, map2);
//		
//		LogReplayAnalyser replayAnalyser = new LogReplayAnalyser(hierarchicalNet, false, initMarking, prlog);
//		LogReplayAnalysisResults lrar = replayAnalyser.AnalyseLogReplay(context); //analyse all traces available
//		TraceReplayCollection provider2 = lrar.trc;
//		
////		PSENLogReplayConnection lrc = new PSENLogReplayConnection(pgexp, provider2);
////		context.addConnection(lrc);
////		
////		TraceReplayCollection trc = provider2.ProjectBack(pgflat, map2, prlog);
////
////		PSENLogReplayConnection lrc2 = new PSENLogReplayConnection(pgflat, trc);
////		context.addConnection(lrc2);
////		PSENLogReplayConnection lrc3 = new PSENLogReplayConnection(hierarchicalNet, trc);
////		context.addConnection(lrc3);
//		
//		provider2.addSubNetEvents(new PetriNetDataSource(hierarchicalNet));
//
//		//create and fill PSEN data structure		
//		PSENProvider provider = new PSENProvider(context, hierarchicalNet, prlog, provider2);
//		
//		return provider;
//	}
	private PSENProvider ReplayOnHierarchicalExpandedNet(final UIPluginContext context, Petrinet hierarchicalNet, Petrinet pgflat, Petrinet pgexp, PetriNet1toManyMapping map2, Marking initMarking, XLog prlog)
	{
		XLog rlog = ReplayToLogConverter.Project(prlog, map2);

		PetriNetDataSource pnds = new PetriNetDataSource(pgexp);
		LogReplayAnalyser replayAnalyser = new LogReplayAnalyser(pnds, true, initMarking, rlog);
		LogReplayAnalysisResults lrar = replayAnalyser.AnalyseLogReplay(context); //analyse all traces available
		TraceReplayCollection provider2 = lrar.trc;
		
		PSENLogReplayConnection lrc = new PSENLogReplayConnection(pgexp, prlog, provider2);
		context.addConnection(lrc);
		
		TraceReplayCollection trc = provider2.ProjectBack(pgflat, map2, prlog);

		PSENLogReplayConnection lrc2 = new PSENLogReplayConnection(pgflat, prlog, trc);
		context.addConnection(lrc2);
		PSENLogReplayConnection lrc3 = new PSENLogReplayConnection(hierarchicalNet, prlog, trc);
		context.addConnection(lrc3);
		
		trc.addSubNetEvents(new PetriNetDataSource(hierarchicalNet));

		//create and fill PSEN data structure		
		PSENProvider provider = new PSENProvider(context, hierarchicalNet, prlog, trc);
		
		return provider;
	}
}
