package org.processmining.processprofiler3d.psen.plugins;


import java.awt.Dimension;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JLabel;
import javax.swing.JPanel;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.processprofiler3d.compviz.PetriNetDataSource;
import org.processmining.processprofiler3d.psen.AlignedLog;
import org.processmining.processprofiler3d.psen.LogAnalyser;
import org.processmining.processprofiler3d.psen.PSENProvider;
import org.processmining.processprofiler3d.psen.connections.PSENConnection;
import org.processmining.processprofiler3d.psen.datastructure.IPerformanceStatisticsEnrichedNet;
import org.processmining.processprofiler3d.psen.gui.SlickerOptionPane;
import org.processmining.processprofiler3d.psen.replayanalysis.LogReplayAnalyser;
import org.processmining.processprofiler3d.psen.replayanalysis.LogReplayAnalysisResults;
import org.processmining.processprofiler3d.psen.replayanalysis.TraceReplayCollection;


/*
 * This is the entry class of the comparative visualisation plug-in.
 * */
public class ModelComparison3 extends PP3DPluginBase
{
	@Plugin(
		name = "Process Profiler 3D: Generate Performance Profile 3", 
		parameterLabels = { "Petri Net", "Performance Aligned Event Log"}, 
		returnLabels = { "Performance Statistics Enriched Net" }, 
		returnTypes = { IPerformanceStatisticsEnrichedNet.class}, 
		userAccessible = false, 
		help = "This plug-in compares the differences between cohorts in an aligned process log."
	)
	@UITopiaVariant(
		affiliation = "Queensland University of Technology", 
		author = "Erik Poppe", 
		email = "erik.poppe@qut.edu.au"
	)

	public IPerformanceStatisticsEnrichedNet replayALog(final UIPluginContext context, Petrinet hierarchicalNet, XLog log) 
	{
		LogAnalyser la = new LogAnalyser(log);
		
		//check if log is aligned log
		if (!AlignedLog.CheckLogIsAligned(la))
		{
			//SlickerOptionPane.showMessageDialog(new JPanel(), "Log is not aligned! Cannot analyse.");
			SlickerOptionPane.showErrorMessage(context, "Log is not aligned! Cannot analyse.", "Error");
			System.out.println("Log is not aligned! Cannot analyse.");
			context.getFutureResult(0).cancel(true);
			return null;
		}
		
		//get initial markings
		Marking initMarking = getInitialMarking(context, hierarchicalNet);
		if (initMarking == null)
		{
			//SlickerOptionPane.showMessageDialog(new JPanel(), "Initial Marking for petrinet could not be found or created! Cannot simulate replay.");
			SlickerOptionPane.showErrorMessage(context, "Initial Marking for petrinet could not be found or created! Cannot simulate replay.", "Error");
			System.out.println("Initial Marking for petrinet could not be found or created! Cannot simulate replay.");

			context.getFutureResult(0).cancel(true);
			return null;
		}
		
		if (initMarking.isEmpty())
		{
			//SlickerOptionPane.showMessageDialog(new JPanel(), "Initial Marking for petrinet is empty! Cannot simulate replay.");
			SlickerOptionPane.showErrorMessage(context, "Initial Marking for petrinet is empty! Cannot simulate replay.", "Error");
			System.out.println("Initial Marking for petrinet is empty! Cannot simulate replay.");

			context.getFutureResult(0).cancel(true);
			return null;
		}
		
		//check net structure
		PetriNetDataSource pnds = new PetriNetDataSource(hierarchicalNet);
		ModelStructureChecker msc = new ModelStructureChecker(pnds);
		if (!msc.isStructureCompliant())
		{
			JPanel listPane = GetErrorPanel(msc);
			InteractionResult ir = context.showConfiguration("Error", listPane);
			if (ir == InteractionResult.CANCEL)
			{
				context.getFutureResult(0).cancel(true);
				return null;
			}
		}
		
		Object[] results = null;
		PSENProvider provider = null;
		
		provider = ReplayOnHierarchicalExpandedNet(context, hierarchicalNet, initMarking, log);
		
		if (provider != null)
		{
			IPerformanceStatisticsEnrichedNet psen = provider.getPSEnet();
			IPerformanceStatisticsEnrichedNet wten = provider.getWaitingTimeTree();
			
			if (psen != null && wten != null)
			{
				//add waitingtimeinfoconnection to context
				context.addConnection(new PSENConnection(psen, wten));
	
				System.out.println("PSEN value count: " + psen.getValueCount());
				System.out.println("WTEN value count: " + wten.getValueCount());
				
				results = new Object[1];
				results[0] = provider.getPSEnet();	
			}
			else
			{
				context.getFutureResult(0).cancel(true);
			}
		}	
		else
		{
			context.getFutureResult(0).cancel(true);
		}
		return (IPerformanceStatisticsEnrichedNet)results[0];
	}

	private JPanel GetErrorPanel(ModelStructureChecker msc)
	{
		JPanel listPane = new JPanel();
		listPane.setLayout(new BoxLayout(listPane, BoxLayout.PAGE_AXIS));
		
		String errorMsg = "Petri net does not comply to structural requirements:";

		JLabel lbl = new JLabel(errorMsg);
		lbl.setSize(100, 500);
		
		listPane.add(lbl);
		
		for (String sx : msc.problemList)
		{
			errorMsg += " \r\n" + sx;
			JLabel lblC = new JLabel("- " + sx);
			listPane.add(lblC);
		}
		
		listPane.add(Box.createRigidArea(new Dimension(5,30)));

		JLabel lblY = new JLabel("Analysis might be flawed.");
		listPane.add(lblY);
		
		JLabel lblX = new JLabel("Do you want to continue anyway?");
		listPane.add(lblX);
		
		System.out.println(errorMsg);
		return listPane;
	}
	
	private PSENProvider ReplayOnHierarchicalExpandedNet(final UIPluginContext context, Petrinet hierarchicalNet, Marking initMarking, XLog prlog)
	{
		TraceReplayCollection trc = GenerateTraceReplayCollection(context, new PetriNetDataSource(hierarchicalNet), initMarking, prlog);

		//create and fill PSEN data structure		
		PSENProvider provider = new PSENProvider(context, hierarchicalNet, prlog, trc);
		
		return provider;
	}

	public static TraceReplayCollection GenerateTraceReplayCollection(final UIPluginContext context, PetriNetDataSource pnds, Marking initMarking,
			XLog prlog) 
	{
		LogReplayAnalyser replayAnalyser = new LogReplayAnalyser(pnds, false, initMarking, prlog);
		LogReplayAnalysisResults lrar = replayAnalyser.AnalyseLogReplay(context); //analyse all traces available
		TraceReplayCollection trc = lrar.trc;
		
		trc.addSubNetEvents(pnds);
		return trc;
	}
}
