package org.processmining.processprofiler3d.psen.replayanalysis.tokenflow;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.processmining.models.graphbased.LocalNodeID;
import org.processmining.models.graphbased.directed.petrinet.elements.Arc;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.processprofiler3d.compviz.INetDataSource;
import org.processmining.processprofiler3d.compviz.SubNetDataSource;
import org.processmining.processprofiler3d.psen.replayanalysis.ITraceAlignment;
import org.processmining.processprofiler3d.psen.replayanalysis.TraceReplayStepData;
import org.processmining.processprofiler3d.psen.replayanalysis.TransitionEvent;

public class TokenFlowSimulation 
{
	private INetDataSource newNet;
	private Map<LocalNodeID, Vector<TokenFlowToken>> mapTokens = new HashMap<LocalNodeID, Vector<TokenFlowToken>>();
	private Map<LocalNodeID, Vector<TransitionEvent>> tEvents = new HashMap<LocalNodeID, Vector<TransitionEvent>>();
	//private Vector<TransitionEvent> tEventsLst = new Vector<TransitionEvent>();
	private Map<LocalNodeID, INetDataSource> tMap = new HashMap<LocalNodeID, INetDataSource>();
	
	private TokenFlowSimulation(INetDataSource pnet)
	{
		newNet = pnet;
	}
	
	public TokenFlowSimulation(INetDataSource pnet, Marking mNewNet)
	{
		newNet = pnet;
		Place[] placeArray = getPlaceArray(pnet);
		
		//create initial tokens
		for (int i = 0; i < placeArray.length; i++) 
		{
			int num = mNewNet.occurrences(placeArray[i]).shortValue();
			for (int j = 0; j < num; j++)
			{
				TokenFlowToken rt = new TokenFlowToken(null);
				addToPlace(placeArray[i], rt);
			}			
		}
	}
	
	public TokenFlowSimulation clone()
	{
		TokenFlowSimulation clonedMarking = new TokenFlowSimulation(newNet);
		
		for (LocalNodeID lid : mapTokens.keySet())
		{
			for (TokenFlowToken rt : mapTokens.get(lid))
			{
				clonedMarking.addToPlace(lid, rt.clone());
			}
		}
		
		return clonedMarking;
	}
	
	public void InitResults()
	{	
		//intialize transitionevent map
		InitResultsHelper(newNet);
	}
	
	private void InitResultsHelper(INetDataSource nds)
	{
		for (Transition t : nds.getTransitions())
		{
			tEvents.put(t.getLocalID(), new Vector<TransitionEvent>());
			tMap.put(t.getLocalID(), nds);
		}
		
		for (SubNetDataSource snds : nds.getSubnets())
		{
			InitResultsHelper(snds);
		}
	}
	
	public TokenFlowResults getTransitionEvents()
	{
		return new TokenFlowResults(tEvents);
	}

	//create an array of all places in net, sorted alphabetically
	private Place[] getPlaceArray(INetDataSource newNet) 
	{
		// place array
		List<Place> placeList = new ArrayList<Place>(newNet.getPlaces());
		Collections.sort(placeList, new Comparator<Place>()
		{
			public int compare(Place o1, Place o2) 
			{
				return o1.getLabel().compareTo(o2.getLabel());
			}
		});
		
		return placeList.toArray(new Place[placeList.size()]);
	}
	
	private Vector<TokenFlowToken> get(Place p)
	{
		return get(p.getLocalID());
	}
	
	private Vector<TokenFlowToken> get(LocalNodeID pId)
	{
		if (!mapTokens.containsKey(pId))
		{
			mapTokens.put(pId, new Vector<TokenFlowToken>());
		}
		return mapTokens.get(pId);
	}
	
	private int getNumTokens(Place p)
	{
		return getNumTokens(p.getLocalID());
	}
	
	private int getNumTokens(LocalNodeID pId)
	{
		if (mapTokens.containsKey(pId))
		{
			return mapTokens.get(pId).size();
		}
		return 0;
	}
	
	private void addToPlace(Place p, TokenFlowToken rt)
	{
		addToPlace(p.getLocalID(), rt);
	}
	
	private void addToPlace(LocalNodeID pId, TokenFlowToken rt)
	{
		if (!mapTokens.containsKey(pId))
		{
			mapTokens.put(pId, new Vector<TokenFlowToken>());
		}
		
		mapTokens.get(pId).add(rt);
	}
	
	public boolean bTransitionFired = false;
	public boolean bTokenMismatch = false;
	private Vector<TokenFlowToken> parentTokens = new Vector<TokenFlowToken>();
	private Vector<TokenFlowToken> childTokens = new Vector<TokenFlowToken>();
	
	public TransitionEvent fireTransition(Transition trans, ITraceAlignment ta) 
	{	
		TransitionEvent lastEvent = new TransitionEvent(trans);
		
		bTransitionFired = true;
		bTokenMismatch = false;
		
		String elog = "";
		parentTokens.clear();
		childTokens.clear();
		
		Collection<ITokenFlowArc> edges = getInArcs(trans);
		for (ITokenFlowArc edge : edges) 
		{
			Place pSource = (Place)edge.getSource();
			
			if (edge.isResetArc()) 
			{
				Vector<TokenFlowToken> v = get(pSource);
				for (int i = 0; i < v.size(); i++)
				{
					TokenFlowToken rt = v.get(0);
					parentTokens.add(rt);
					rt.consume(lastEvent);
					
					v.remove(0);
				}

				elog = "ResetArc happened!";
			} 
			else if (!(edge.isInhibitorArc())) 
			{
				// ordinary edges
				int removedTokens = edge.getWeight();
				int numTokensAvailable = getNumTokens(pSource);
				int numAfter = (short)Math.max(0, (numTokensAvailable - removedTokens));

				Vector<TokenFlowToken> v = get(pSource);
				//remove existing tokens
				for (int i = 0; i < (numTokensAvailable - numAfter); i++)
				{
					TokenFlowToken rt = v.get(0);
					parentTokens.add(rt);
					rt.consume(lastEvent);
					v.remove(0);
				}
				if ((removedTokens - (numTokensAvailable - numAfter)) > 0)
				{
					bTokenMismatch = true;
				}
				
				elog = "NonInhibitorArc happened! Weight:" + removedTokens + " Available:" + numTokensAvailable + " numAfter:" + numAfter;
			}
			else
			{
				elog = "Other happened!";
			}
		}
		
		edges = getOutArcs(trans);
		for (ITokenFlowArc edge : edges) 
		{
			int addedTokens = edge.getWeight();
			Place pTarget = (Place)edge.getTarget();
			
			for (int i = 0; i < addedTokens; i++)
			{
				TokenFlowToken rt = new TokenFlowToken(lastEvent);
				childTokens.add(rt);
				addToPlace(pTarget, rt);
			}
		}
		
		if (lastEvent.getParents().isEmpty() && lastEvent.getScheduleId() > 0)
		{
			System.out.println(elog);
			System.out.println("No parent! " + lastEvent.getTransitionLabel());
		}

		tEvents.get(trans.getLocalID()).add(lastEvent);
		//tEventsLst.add(lastEvent);
		
		return lastEvent;
	}
	
	public TransitionEvent fireTransition(Transition trans, ITraceAlignment ta, TraceReplayStepData trs) 
	{	
		TransitionEvent lastEvent = fireTransition(trans, ta);
		
		trs.bTransitionFired = bTransitionFired;
		trs.bTokenMismatch = bTokenMismatch;
		trs.parentTokens = parentTokens;
		trs.childTokens = childTokens;
		
		return lastEvent;
	}
	
	private Collection<ITokenFlowArc> getInArcs(Transition trans)
	{	
		//System.out.println("getInArcs!");
		INetDataSource inds = tMap.get(trans.getLocalID());
		Place startPlace = inds.getStartPlace();
		
		Collection<Arc> edges = inds.getInEdges(trans);
		Vector<ITokenFlowArc> lst = new Vector<ITokenFlowArc>();
		for (Arc edge : edges)
		{
			if (edge.getSource() == startPlace && inds.hasParent())
			{
				INetDataSource pnds = inds.getParent();
				Collection<Arc> edges2 = pnds.getInEdges(inds.getParentTrans());
				for (Arc edge2 : edges2)
				{
					TokenFlowArcCL tfa = new TokenFlowArcCL(edge2.getSource(), edge.getTarget(), 1);
					lst.add(tfa);
				}
			}
			else
			{
				TokenFlowArc tfa = new TokenFlowArc(edge);
				lst.add(tfa);
			}
		}
		return lst;
	}
	
	private Collection<ITokenFlowArc> getOutArcs(Transition trans)
	{
		//System.out.println("getOutArcs!");
		INetDataSource inds = tMap.get(trans.getLocalID());
		Place endPlace = inds.getEndPlace();
		
		Collection<Arc> edges = inds.getOutEdges(trans);
		Vector<ITokenFlowArc> lst = new Vector<ITokenFlowArc>();
		for (Arc edge : edges)
		{
			if (edge.getTarget() == endPlace && inds.hasParent())
			{
				INetDataSource pnds = inds.getParent();
				Collection<Arc> edges2 = pnds.getOutEdges(inds.getParentTrans());
				for (Arc edge2 : edges2)
				{
					TokenFlowArcCL tfa = new TokenFlowArcCL(edge.getSource(), edge2.getTarget(), 1);
					lst.add(tfa);
				}
			}
			else
			{
				TokenFlowArc tfa = new TokenFlowArc(edge);
				lst.add(tfa);
			}
		}
		return lst;
	}
	
	public String getStateDesc()
	{
		String str = "";
		
		for (LocalNodeID lid : mapTokens.keySet())
		{
			Vector<TokenFlowToken> lst = mapTokens.get(lid);
			if (!lst.isEmpty())
			{
				for (Place p : newNet.getPlaces())
				{
					if (p.getLocalID() == lid)
					{
						str += "" + p.getLabel() + "(" + lst.size() + "), ";
					}
				}
			}
		}
		
		return str;
	}
}
